###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

from PyQt4 import QtCore, QtGui
from PyQt4.QtCore import pyqtSignal, pyqtSlot
import vistrails.core.system
from vistrails.gui.common_widgets import QDockPushButton
from vistrails.gui.mashups.mashups_manager import MashupsManager
from vistrails.gui.utils import show_warning
from vistrails.gui.vistrails_palette import QVistrailsPaletteInterface

class QMashupsInspector(QtGui.QFrame, QVistrailsPaletteInterface):
    """
    QMashupsInspector is a widget with tabs showing properties of the selected
    mashuptrail. It contains a list of tagged mashups, and when a mashup is
    selected, it shows the list of aliases in the mashup.
    
    """
    #signals
    mashupChanged = pyqtSignal()
    def __init__(self, controller=None, parent=None):
        """ QMashupsInspector(controller: MashupController,
                            parent: QWidget) -> QMashupsInspector
        
        """
        QtGui.QFrame.__init__(self, parent)
        QVistrailsPaletteInterface.__init__(self)
        #print "****** Inspector INIT"
        self.set_title("Mashups Inspector")
        self.setFrameStyle(QtGui.QFrame.Panel|QtGui.QFrame.Sunken)
        self.setSizePolicy(QtGui.QSizePolicy.Expanding,
                           QtGui.QSizePolicy.Expanding)
        
        #ok, this will store the original vistrail controller
        self.controller = controller
        self.mshpController = None #MashupController
        self.manager = MashupsManager.getInstance()
        
        layout = QtGui.QVBoxLayout()
        layout.setMargin(2)
        layout.setSpacing(3)
        self.workflowLabel = QtGui.QLabel("Workflow: ")
        layout.addWidget(self.workflowLabel)
        self.mashupsList = QMashupsListPanel(parent=self)
        listLabel = QtGui.QLabel("Available Mashups:")
        layout.addWidget(listLabel)
        layout.addWidget(self.mashupsList)
        
        self.mashupInspector = QMashupProp(parent=self)
        inspector_group = QtGui.QGroupBox("Mashup properties")
        g_layout = QtGui.QVBoxLayout()
        g_layout.setMargin(1)
        g_layout.setSpacing(3)
        g_layout.addWidget(self.mashupInspector)
        inspector_group.setLayout(g_layout)
        layout.addWidget(inspector_group)
        self.setLayout(layout)
                    
    def updateVistrailController(self, controller):
        self.controller = controller
        #print "         *** Mashup Inspector: controller changed ", controller
    
    def updateVistrailVersion(self, version):
        if self.controller:
            self.vt_version = version
                
        #print "         *** Mashup Inspector: version changed ", version
    
    def updateMshpController(self, mshpController):
        #print "     **** updateMshpController", mshpController
        if (self.mshpController is not None and 
            self.mshpController != mshpController):
            self.mshpController.stateChanged.disconnect(self.stateChanged)
        self.mshpController = mshpController
        self.workflowLabel.setText(
          "Workflow: <b>%s</b>"%self.mshpController.getVistrailWorkflowTag())
        self.mshpController.stateChanged.connect(self.stateChanged)
        self.mashupInspector.updateController(mshpController)
        self.mashupsList.updateController(mshpController)
        
    def updateMshpVersion(self, version):
        pass
        #print "updateMshpVersion", version
        
    def stateChanged(self):
        versionId = self.mshpController.currentVersion
        self.mashupInspector.updateVersion(versionId)
        self.mashupsList.stateChanged()
        
################################################################################        

class QMashupProp(QtGui.QWidget):
    def __init__(self, controller=None, parent=None):
        QtGui.QWidget.__init__(self, parent)
        self.controller = controller
        self.versionNumber = -1
        vLayout = QtGui.QVBoxLayout()
        vLayout.setMargin(0)
        vLayout.setSpacing(5)
        self.setLayout(vLayout)
        gLayout = QtGui.QGridLayout()
        gLayout.setMargin(0)
        gLayout.setSpacing(5)
        gLayout.setColumnMinimumWidth(1,5)
        gLayout.setRowMinimumHeight(0,24)
        gLayout.setRowMinimumHeight(1,24)
        gLayout.setRowMinimumHeight(2,24)
        gLayout.setRowMinimumHeight(3,24)        
        vLayout.addLayout(gLayout)
        
        #vtVersionLabel = QtGui.QLabel('Workflow:', self)
        #gLayout.addWidget(vtVersionLabel, 0, 0, 1, 1)
        
        #self.vtVersionEdit = QtGui.QLabel('', self)
        #gLayout.addWidget(self.vtVersionEdit, 0, 2, 1, 1)
        
        tagLabel = QtGui.QLabel('Mashup Tag:', self)
        gLayout.addWidget(tagLabel, 0, 0, 1, 1)

        editLayout = QtGui.QHBoxLayout()
        editLayout.setMargin(0)
        editLayout.setSpacing(2)
        self.tagEdit = QtGui.QLineEdit()
        tagLabel.setBuddy(self.tagEdit)
        editLayout.addWidget(self.tagEdit)
        self.tagEdit.setEnabled(False)

        self.tagReset = QtGui.QToolButton(self)
        self.tagReset.setIcon(QtGui.QIcon(
                self.style().standardPixmap(QtGui.QStyle.SP_DialogCloseButton)))
        self.tagReset.setIconSize(QtCore.QSize(12,12))
        self.tagReset.setAutoRaise(True)
        self.tagReset.setEnabled(False)
        editLayout.addWidget(self.tagReset)

        gLayout.addLayout(editLayout, 0, 2, 1, 1)

        userLabel = QtGui.QLabel('User:', self)
        gLayout.addWidget(userLabel, 1, 0, 1, 1)
        
        self.userEdit = QtGui.QLabel('', self)
        gLayout.addWidget(self.userEdit, 1, 2, 1, 1)

        dateLabel = QtGui.QLabel('Date:', self)
        gLayout.addWidget(dateLabel, 2, 0, 1, 1)

        self.dateEdit = QtGui.QLabel('', self)
        gLayout.addWidget(self.dateEdit, 2, 2, 1, 1)
        
        self.btnExport = QDockPushButton("Export...")
        gLayout.addWidget(self.btnExport, 3,0,1,3, QtCore.Qt.AlignHCenter)
        
        vLayout.addStretch()
        
        self.connect(self.tagEdit, QtCore.SIGNAL('editingFinished()'),
                     self.tagFinished)
        self.connect(self.tagEdit, QtCore.SIGNAL('textChanged(QString)'),
                     self.tagChanged)
        self.connect(self.tagReset, QtCore.SIGNAL('clicked()'),
                     self.tagCleared)
        self.connect(self.btnExport, QtCore.SIGNAL("clicked()"),
                     self.exportMashupGUI)
        
    def updateController(self, mshpController):
        self.controller = mshpController
       
        #print "QMashupProp.updateController ", self.controller, self.controller.currentVersion
        if self.controller and self.controller.currentVersion > -1:
            self.versionNumber = self.controller.currentVersion
            self.tagEdit.setText(self.controller.mshptrail.getTagForActionId(
                                            self.controller.currentVersion))
            self.tagEdit.setEnabled(True)
            action = self.controller.mshptrail.actionMap[self.controller.currentVersion]
            self.userEdit.setText(action.user)
            self.dateEdit.setText(action.date)
            #self.vtEdit.setText(self.controller.getVistrailName())
            #self.vtVersionEdit.setText(self.controller.getVistrailWorkflowTag())
            return
        else:
            self.tagEdit.setEnabled(False)
            self.tagReset.setEnabled(False)
            self.tagEdit.setText('')
            self.userEdit.setText('')
            self.dateEdit.setText('')  
            #self.vtEdit.setText('')
            #self.vtVersionEdit.setText('')
        
    def updateVersion(self, versionNumber):
        self.versionNumber = versionNumber
        if self.controller and self.versionNumber > -1:
            tagtext = self.controller.mshptrail.getTagForActionId(
                                    self.versionNumber)
            #print "updateVersion", versionNumber, tagtext
            self.tagEdit.setText(tagtext)
            action = self.controller.mshptrail.actionMap[self.versionNumber]
            self.userEdit.setText(action.user)
            self.dateEdit.setText(action.date)
            #self.vtEdit.setText(self.controller.getVistrailName())
            #self.vtVersionEdit.setText(self.controller.getVistrailWorkflowTag())
        else:
            self.tagEdit.setEnabled(False)
            self.tagReset.setEnabled(False)
            self.tagEdit.setText('')
            self.userEdit.setText('')
            self.dateEdit.setText('')  
            #self.vtEdit.setText('')
            #self.vtVersionEdit.setText('')
            
    def tagFinished(self):
        """ tagFinished() -> None
        Update the new tag to mashup
        
        """
        if self.controller:
            name = self.controller.mshptrail.getTagForActionId(self.versionNumber)
            currentText = str(self.tagEdit.text())
            if name != currentText:    
                #print "will update current tag", currentText
                self.controller.updateCurrentTag(currentText)
                
                
    def tagChanged(self, text):
        """ tagChanged(text: QString) -> None
        Update the button state if there is text

        """
        self.tagReset.setEnabled(text != '')

    def tagCleared(self):
        """ tagCleared() -> None
        Remove the tag
        
        """ 
        self.tagEdit.setText('')
        self.tagFinished()
        
    def exportMashupGUI(self):
        if self.controller:
            dialog = QMashupExportDialog(self)
            if dialog.exec_() == QtGui.QDialog.Accepted:
                result = dialog.btnPressed
                fileName = QtGui.QFileDialog.getSaveFileName(
                           self,
                           "Export Mashup...",
                           vistrails.core.system.vistrails_file_directory(),
                           "VisTrail link files (*.vtl)")
                if fileName:
                    filename = str(fileName)
                    res = MashupsManager.exportMashup(filename, 
                                                      self.controller.originalController, 
                                                      self.controller.mshptrail,
                                                      self.controller.currentVersion,
                                                      result)
                    if not res:
                        show_warning('VisTrails - Exporting Mashup',
                    'There was an error and the mashup could not be exported.')
                        
################################################################################
            
class QMashupsListPanel(QtGui.QWidget):
    def __init__(self, controller=None, parent=None):
        QtGui.QWidget.__init__(self, parent)
        self.controller = controller
        self.mashupsList = QtGui.QListWidget()
        self.mashupsList.setSelectionMode(QtGui.QAbstractItemView.SingleSelection)
        self.mashupsList.itemSelectionChanged.connect(self.changeSelection)
        layout = QtGui.QVBoxLayout()
        layout.setMargin(0)
        layout.setSpacing(5)
        layout.addWidget(self.mashupsList)
        self.setLayout(layout)
            
    def updateController(self, mshpController):
        self.controller = mshpController
        self.stateChanged()
                    
    def stateChanged(self):
        self.mashupsList.itemSelectionChanged.disconnect(self.changeSelection)
        self.mashupsList.clear()
        if self.controller:
            self.tagMap = self.controller.mshptrail.getTagMap()
            currentTag = self.controller.getCurrentTag()
            tags = self.tagMap.keys()
            latestversion = self.controller.mshptrail.getLatestVersion()
            if not self.controller.versionHasTag(latestversion):
                item = QMashupListPanelItem("<latest>",
                                            latestversion,
                                            self.mashupsList)
                if latestversion == self.controller.currentVersion:
                    item.setSelected(True)

            if len(tags) > 0:
                tags.sort()
                for tag in tags:
                    item = QMashupListPanelItem(str(tag),
                                                self.tagMap[tag],
                                                self.mashupsList)
                    if tag == currentTag:
                        item.setSelected(True)
                        
        self.mashupsList.itemSelectionChanged.connect(self.changeSelection)
    
    @pyqtSlot()
    def changeSelection(self):
        items = self.mashupsList.selectedItems()
        if len(items) == 1:
            version = items[0].version
            if version != self.controller.currentVersion:
                self.controller.setCurrentVersion(version, quiet=False)
            
    def selectMashup(self, name):
        itemlist = self.mashupsList.findItems(name,
                                              QtCore.Qt.MatchExactly)
        if len(itemlist) == 1:
            item = itemlist[0]
            self.mashupsList.setCurrentItem(item)
            
################################################################################

class QMashupListPanelItem(QtGui.QListWidgetItem):
    def __init__(self, tag, version, parent=None):
        QtGui.QListWidgetItem.__init__(self, tag, parent)
        self.tag = tag
        self.version = version
        
################################################################################

class QMashupExportDialog(QtGui.QDialog):
    FULLTREE = 0
    MINIMAL = 1
    LINK = 2
    def __init__(self, parent=None):
        QtGui.QDialog.__init__(self, parent)
        self.setWindowTitle('VisTrails - Exporting Mashup')
        dlgLayout = QtGui.QVBoxLayout()
        gb = QtGui.QGroupBox("This will export the mashup as a file:")
        gblayout = QtGui.QVBoxLayout()
        self.rbMinimal = QtGui.QRadioButton("Include only this mashup\
 and its original workflow")
        self.rbMinimal.setChecked(True)
        self.rbFullTree = QtGui.QRadioButton("Include full tree (this will also\
 include other mashups)")
        self.rbLink = QtGui.QRadioButton("As a link (this will work only on\
 this machine)")
        gblayout.addWidget(self.rbMinimal)
        gblayout.addWidget(self.rbFullTree)
        gblayout.addWidget(self.rbLink)
        gb.setLayout(gblayout)
        
        btnOk = QtGui.QPushButton("OK")
        btnCancel = QtGui.QPushButton("Cancel")
        btnLayout = QtGui.QHBoxLayout()
        btnLayout.addStretch()
        btnLayout.addWidget(btnOk)
        btnLayout.addWidget(btnCancel)
        btnLayout.addStretch()
        
        dlgLayout.addWidget(gb)
        dlgLayout.addLayout(btnLayout)
        self.setLayout(dlgLayout)
        
        self.btnPressed = -1
        self.connect(btnOk, QtCore.SIGNAL("clicked()"), self.btnOkPressed) 
        self.connect(btnCancel, QtCore.SIGNAL("clicked()"), self.btnCancelPressed)
        
    def btnOkPressed(self):
        if self.rbFullTree.isChecked():
            self.btnPressed = self.FULLTREE
        elif self.rbMinimal.isChecked():
            self.btnPressed = self.MINIMAL
        elif self.rbLink.isChecked():
            self.btnPressed = self.LINK
        self.accept()
        
    def btnCancelPressed(self):
        self.btnPressed = -1
        self.reject()


#!/usr/bin/env python

# This simple example shows how to do basic rendering and pipeline
# creation.

import vtk
# The colors module defines various useful colors.
from vtk.util.colors import tomato

# This creates a polygonal cylinder model with eight circumferential
# facets.
cylinder = vtk.vtkCylinderSource()
cylinder.SetResolution(8)

# The mapper is responsible for pushing the geometry into the graphics
# library. It may also do color mapping, if scalars or other
# attributes are defined.
cylinderMapper = vtk.vtkPolyDataMapper()
cylinderMapper.SetInputConnection(cylinder.GetOutputPort())

# The actor is a grouping mechanism: besides the geometry (mapper), it
# also has a property, transformation matrix, and/or texture map.
# Here we set its color and rotate it -22.5 degrees.
cylinderActor = vtk.vtkActor()
cylinderActor.SetMapper(cylinderMapper)
cylinderActor.GetProperty().SetColor(tomato)
cylinderActor.RotateX(30.0)
cylinderActor.RotateY(-45.0)

# Create the graphics structure. The renderer renders into the render
# window. The render window interactor captures mouse events and will
# perform appropriate camera or actor manipulation depending on the
# nature of the events.
ren = vtk.vtkRenderer()
renWin = vtk.vtkRenderWindow()
renWin.AddRenderer(ren)
iren = vtk.vtkRenderWindowInteractor()
iren.SetRenderWindow(renWin)

# Add the actors to the renderer, set the background and size
ren.AddActor(cylinderActor)
ren.SetBackground(0.1, 0.2, 0.4)
renWin.SetSize(200, 200)

# This allows the interactor to initalize itself. It has to be
# called before an event loop. 
iren.Initialize()

# We'll zoom in a little by accessing the camera and invoking a "Zoom"
# method on it.
ren.ResetCamera()
ren.GetActiveCamera().Zoom(1.5)
renWin.Render()

# Start the event loop.
iren.Start()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
""" Contains classes for persisting running executions to disk

"""

from __future__ import division

from vistrails.core.configuration import get_vistrails_configuration
from vistrails.core import debug
from vistrails.core.modules.vistrails_module import NotCacheable, \
    ModuleError, ModuleSuspended

from uuid import uuid1

import datetime
import getpass
import json
import time
import unittest
import weakref


class JobMixin(NotCacheable):
    """ Mixin for suspendable modules.

    This provides the base behavior for modules that submit jobs by handling
    the serialization & JobMonitor interaction for you.

    The module developer needs only implement the following methods:
        job_read_inputs()
        job_set_results()
        job_start()
        job_get_handle()
        job_finish()
    """

    def compute(self):
        """ Base behavior for job-submitting modules.

        This provides the base code and calls the methods that the module
        developer should provide.
        """
        debug.log("%s compute() starting\n"
                  "signature = %r" % (self.__class__.__name__,
                                      self.signature))

        jm = self.job_monitor()

        cache = jm.getCache(self.signature)
        if cache is not None:
            # Result is available from cache
            jm.setCache(self.signature, cache.parameters)
            debug.log("Cached results found; calling job_set_results()")
            self.job_set_results(cache.parameters)
            return
        else:
            debug.log("Cache miss")

        job = jm.getJob(self.signature)
        if job is None:
            debug.log("Job doesn't exist")
            params = self.job_read_inputs()
            params = self.job_start(params)
        else:
            debug.log("Got job from JobMonitor")
            params = job.parameters
        jm.addJob(self.signature, params, self.job_name())

        # Might raise ModuleSuspended
        debug.log("Calling checkJob()")
        try:
            jm.checkJob(self, self.signature, self.job_get_handle(params))
        except ModuleSuspended, e:
            debug.log("checkJob() raised ModuleSuspended, job handle is %r" %
                      e.handle)
            raise

        # Didn't raise: job is finished
        debug.log("Calling job_finish()")
        params = self.job_finish(params)
        debug.log("Filling cache")
        jm.setCache(self.signature, params)
        debug.log("Calling job_set_results()")
        self.job_set_results(params)

    def update_upstream(self):
        """ Decides whether or not to run the upstream.

        If a job has already been submitted and the local JobMonitor knows of
        it, we don't need to run upstream modules to check the status.

        If status check indicates that the job no longer exists, then we should
        run upstream then submit again.
        """
        if not hasattr(self, 'signature'):
            raise ModuleError(self, "Module has no signature")
        jm = self.job_monitor()
        if not (jm.getCache(self.signature) or jm.getJob(self.signature)):
            # We need to submit a new job
            # Update upstream, compute() will need it
            super(JobMixin, self).update_upstream()

    def job_read_inputs(self):
        """ Implemented by modules to read job parameters from input ports.

        Returns the `params` dictionary used by subsequent methods.
        """
        raise NotImplementedError

    def job_start(self, params):
        """ Implemented by modules to submit the job.

        Gets the `params` dictionary and returns a new dictionary, for example
        with additional info necessary to check the status later.
        """
        raise NotImplementedError

    def job_finish(self, params):
        """ Implemented by modules to get info from the finished job.

        This is called once the job is finished to get the results. These can
        be added to the `params` dictionary that this method returns.

        This is the right place to clean up the job from the server if they are
        not supposed to persist.
        """
        raise NotImplementedError

    def job_set_results(self, params):
        """ Implemented by modules to set the output ports.

        This is called after job_finished() or after getting the cached results
        to set the output ports on this module, from the `params` dictionary.
        """
        raise NotImplementedError

    def job_get_handle(self, params):
        """ Implemented by modules to return the JobHandle object.

        This returns an object following the JobHandle interface. The
        JobMonitor will use it to check the status of the job and call back
        this module once the job is done.

        JobHandle needs the following method:
          * finished(): returns True if the job is finished
        """
        return None

    def job_name(self):
        """ Readable name for the job.

        Modules needn't override this, in which case a default will be
        provided.
        """
        # use module description if it exists
        if 'pipeline' in self.moduleInfo and self.moduleInfo['pipeline']:
            p_modules = self.moduleInfo['pipeline'].modules
            p_module = p_modules[self.moduleInfo['moduleId']]
            if p_module.has_annotation_with_key('__desc__'):
                return p_module.get_annotation_by_key('__desc__').value
        return self.__class__.__name__


class Workflow(object):
    """ Represents a workflow that has jobs.

    It can have one or several suspended modules.
    It can be serialized to disk.
    """
    def __init__(self, version, name='untitled', id=None, user=None,
                 start=None, jobs=None):
        """ __init__(version: str/int, name: str, id: str,
            user: str, start: str, jobs: list) -> None

            version - workflow version
            name - a human readable name for the job
            id - persistent identifier
            user - who started the job
            start - start time
            jobs - a dict with jobs
        """
        self.version = version
        self.name = name
        self.id = id if id else str(uuid1())
        self.user = getpass.getuser()
        self.start = start if start else datetime.datetime.now().isoformat()
        self.jobs = jobs if jobs else {}
        # parent modules are stored as temporary exceptions
        self.parents = {}

    def to_dict(self):
        wf = dict()
        wf['version'] = self.version
        wf['id'] = self.id
        wf['name'] = self.name
        wf['user'] = self.user
        wf['start'] = self.start
        wf['jobs'] = self.jobs.keys()
        return wf

    @staticmethod
    def from_dict(wf):
        return Workflow(wf['version'], wf['name'], wf['id'],
                        wf['user'], wf['start'], wf['jobs'])

    def __eq__(self, other):
        if self.version != other.version:
            return False
        if self.name != other.name:
            return False
        if self.id != other.id:
            return False
        if self.user != other.user:
            return False
        if self.start != other.start:
            return False
        if len(self.jobs) != len(other.jobs):
            return False
        if set(self.jobs) != set(other.jobs):
            return False
        return True

    def reset(self):
        for job in self.jobs.itervalues():
            job.reset()
        self.parents = {}

    def completed(self):
        """ Returns true if there are no suspended jobs
        """
        for job in self.jobs.itervalues():
            if not job.finished:
                return False
        return True


class Job(object):
    """A suspended module.
    """
    def __init__(self, id, parameters, name='', start=None, finished=False):
        """ __init__(id: str, parameters: dict, name: str, start: str,
                     finished: bool)

            id - persistent identifier
            parameters - either output values or job parameters
            start - start time
            finished - is it finished or running?
        """
        self.id = id
        self.parameters = parameters
        self.name = name
        self.start = start if start else datetime.datetime.now().isoformat()
        self.finished = finished
        self.updated = True

    def reset(self):
        self.updated = False

    def mark(self):
        self.updated = True

    def finish(self, params=None):
        self.parameters = params if params else {}
        self.finished = True

    def description(self):
        return self.parameters.get('__desc__', '')

    def to_dict(self):
        m = dict()
        m['id'] = self.id
        m['parameters'] = self.parameters
        m['name'] = self.name
        m['start'] = self.start
        m['finished'] = self.finished
        return m

    @staticmethod
    def from_dict(m):
        return Job(m['id'],
                   m['parameters'],
                   m['name'],
                   m['start'],
                   m['finished'])

    def __eq__(self, other):
        if self.id != other.id:
            return False
        if self.parameters != other.parameters:
            return False
        if self.start != other.start:
            return False
        if self.finished != other.finished:
            return False
        return True


class JobMonitor(object):
    """ Keeps a list of running jobs and the current job for a vistrail.

    Jobs are added by the interpreter are saved with the vistrail.
    A callback mechanism is used to interact with the associated GUI component.
    """

    def __init__(self, json_string=None):
        self._current_workflow = None
        self.workflows = {}
        self.jobs = {}
        self.callback = None
        if json_string is not None:
            self.unserialize(json_string)

    def setCallback(self, callback=None):
        """ setCallback(callback: class) -> None
            Sets a callback when receiving commands

        """
        self.callback = weakref.ref(callback)

    ###########################################################################
    # Running Workflows

    def serialize(self):
        """ serialize() -> None
            serializes the running jobs to json

        """
        _dict = {}

        jobs = dict()
        for id, job in self.jobs.items():
            jobs[id] = job.to_dict()
        _dict['jobs'] = jobs

        workflows = dict()
        for id, workflow in self.workflows.items():
            workflows[id] = workflow.to_dict()
        _dict['workflows'] = workflows

        return json.dumps(_dict)

    def unserialize(self, s):
        """ unserialize(s: str) -> None
            unserializes the running jobs from json

        """

        _dict = json.loads(s)

        jobs = _dict.get('jobs', {})
        self.jobs = {}
        for id, job in jobs.iteritems():
            self.jobs[id] = Job.from_dict(job)

        workflows = _dict.get('workflows', {})
        self.workflows = {}
        for id, workflow in workflows.iteritems():
            workflow['jobs'] = dict([(i, self.jobs[i])
                                     for i in workflow['jobs']
                                     if i in self.jobs])
            wf = Workflow.from_dict(workflow)
            self.workflows[id] = wf
        return self.workflows

    def addWorkflow(self, workflow):
        """ addWorkflow(workflow: Workflow) -> None

        """
        self.workflows[workflow.id] = workflow
        for id, job in workflow.jobs.iteritems():
            self.jobs[id] = job

    def getWorkflow(self, id):
        """ getWorkflow(id: str) -> Workflow

            Checks if a workflow exists using its id and returns it

        """
        return self.workflows.get(id, None)

    def deleteWorkflow(self, id):
        """ deleteWorkflow(id: str) -> None
            deletes a workflow

        """
        workflow = self.workflows[id]
        del self.workflows[id]
        # delete jobs that only occur in this workflow
        for job_id in workflow.jobs:
            delete = True
            for wf in self.workflows.values():
                if job_id in wf.jobs:
                    delete = False
            if delete:
                del self.jobs[job_id]
        if self.callback is not None and self.callback() is not None:
            self.callback().deleteWorkflow(id)

    def deleteJob(self, id):
        """ deleteJob(id: str, parent_id: str) -> None
            deletes a job from all workflows
        """
        del self.jobs[id]
        for wf in self.workflows.itervalues():
            if id in wf.jobs:
                del wf.jobs[id]
        if self.callback is not None and self.callback() is not None:
            self.callback().deleteJob(id)

    ###########################################################################
    # _current_workflow methods

    def currentWorkflow(self):
        """ currentWorkflow() -> Workflow

        """
        return self._current_workflow

    def startWorkflow(self, workflow):
        """ startWorkflow(workflow: Workflow) -> None

        """
        if self._current_workflow:
            raise Exception("A workflow is still running: %s!" %
                            self._current_workflow)
        workflow.reset()
        self._current_workflow = workflow
        if self.callback is not None and self.callback() is not None:
            self.callback().startWorkflow(workflow)

    def addJobRec(self, obj, parent_id=None):
        workflow = self.currentWorkflow()
        id = obj.module.signature
        if obj.children:
            for child in obj.children:
                self.addJobRec(child, id)
            return
        if id in workflow.jobs:
            # this is an already existing new-style job
            # mark that it has been used
            workflow.jobs[id].mark()
            return
        if id in workflow.jobs:
            # this is an already existing new-style job
            # mark that it has been used
            workflow.jobs[id].mark()
            return
        # this is a new old-style job that we need to add
        self.addJob(id, {'__message__': obj.msg}, obj.name)

    def finishWorkflow(self):
        """ finish_job() -> None

            Finishes the running workflow

        """
        workflow = self._current_workflow
        # untangle parents
        # only keep the top item
        c = set()
        for exception in workflow.parents.itervalues():
            if exception.children:
                c.update([id(child) for child in exception.children])
        for child in c:
            if child in workflow.parents:
                del workflow.parents[child]
        for parent in workflow.parents.itervalues():
            self.addJobRec(parent)

        # Assume all unfinished jobs that were not updated are now finished
        for job in workflow.jobs.values():
            if not job.finished and not job.updated:
                job.finish()
        if self.callback is not None and self.callback() is not None:
            self.callback().finishWorkflow(workflow)
        self._current_workflow = None

    def addJob(self, id, params=None, name='', finished=False):
        """ addJob(id: str, params: dict, name: str, finished: bool) -> uuid

            Adds a job to the currently running workflow

        """

        params = params if params is not None else {}

        if self.hasJob(id):
            # update job attributes
            job = self.getJob(id)
            job.parameters = params
            if name:
                job.name = name
            job.finished = finished
            # we want to keep the start date
        else:
            job = Job(id, params, name, finished=finished)
            self.jobs[id] = job

        workflow = self.currentWorkflow()
        if workflow:
            workflow.jobs[id] = job
            # we add workflows permanently if they have at least one job
            self.workflows[workflow.id] = workflow
        if self.callback is not None and self.callback() is not None:
            self.callback().addJob(self.getJob(id))

    def addParent(self, error):
        """ addParent(id: str, name: str, finished: bool) -> None

            Adds an exception to be used later

        """
        workflow = self.currentWorkflow()
        if not workflow:
            return  # ignore non-monitored jobs
        workflow.parents[id(error)] = error

    def setCache(self, id, params, name=''):
        self.addJob(id, params, name, True)

    def checkJob(self, module, id, handle):
        """ checkJob(module: VistrailsModule, id: str, handle: object) -> None
            Starts monitoring the job for the current running workflow
            module - the module to suspend
            id - the job identifier
            handle - an object following the JobHandle interface, i.e. with a
                finished method for checking if the job has completed

        """
        if not self.currentWorkflow():
            if not handle or not self.isDone(handle):
                raise ModuleSuspended(module, 'Job is running',
                                      handle=handle)
        job = self.getJob(id)
        if self.callback is not None and self.callback() is not None:
            self.callback().checkJob(module, id, handle)
            return

        conf = get_vistrails_configuration()
        interval = conf.jobCheckInterval
        if interval and not conf.jobAutorun:
            if handle:
                # wait for module to complete
                try:
                    while not self.isDone(handle):
                        time.sleep(interval)
                        print ("Waiting for job: %s,"
                               "press Ctrl+C to suspend") % job.name
                except KeyboardInterrupt:
                    raise ModuleSuspended(module, 'Interrupted by user, job'
                                          ' is still running', handle=handle)
        else:
            if not handle or not self.isDone(handle):
                raise ModuleSuspended(module, 'Job is running',
                                      handle=handle)

    def getJob(self, id):
        """ getJob(id: str) -> Job

        """
        return self.jobs.get(id, None)

    def getCache(self, id):
        """ getCache(id: str) -> Job
            Checks if a completed module exists using its id and returns it
        """
        job = self.jobs.get(id, None)
        return job if job and job.finished else None

    def hasJob(self, id):
        """ hasJob(id: str) -> bool

            Checks if a job exists

        """
        return id in self.jobs

    def updateUrl(self, new, old):
        for workflow in self.workflows.values():
            if workflow.vistrail == old:
                workflow.vistrail = new

    def isDone(self, handle):
        """ isDone(self, monitor) -> bool

            A job is done when it reaches finished or failed state
            val() is used by stable batchq branch
        """
        finished = handle.finished()
        if hasattr(finished, 'val'):
            finished = finished.val()
        if finished:
            return True

        # FIXME : deprecate this, remove from RemoteQ
        # finished should just return True here too
        if hasattr(handle, 'failed'):
            failed = handle.failed()
            if hasattr(failed, 'val'):
                failed = failed.val()
            if failed:
                return True
        return False


###############################################################################
# Testing

class TestJob(unittest.TestCase):
    def test_job(self):
        jm = JobMonitor()
        job1 = Job('`13/5', {'a': 3, 'b': '7'})
        job2 = Job('3', {'a': 6}, 'my_name', 'a_string_date', True)
        # test module to/from dict
        job3 = Job.from_dict(job2.to_dict())
        self.assertEqual(job2, job3)

        workflow1 = Workflow(26)
        workflow2 = Workflow('tagname', 'myjob', 'myid', 'tommy',
                             '2013-10-07 13:06',
                             {job1.id: job1, job2.id: job2})
        # test workflow to/from dict
        workflow3 = Workflow.from_dict(workflow2.to_dict())
        self.assertEqual(workflow2, workflow3)

        # test start/finish job
        jm.startWorkflow(workflow2)
        self.assertEqual(workflow2, jm._current_workflow)
        jm.finishWorkflow()
        self.assertEqual(None, jm._current_workflow)

        # test add job
        jm.startWorkflow(workflow2)
        jm.addJob('my_uuid_id', {'myparam': 0})
        self.assertIn('my_uuid_id', workflow2.jobs)
        jm.finishWorkflow()

        # test serialization
        jm.addWorkflow(workflow1)
        jm.addWorkflow(workflow2)
        jm.unserialize(jm.serialize())
        self.assertIn(workflow1.id, jm.workflows)
        self.assertIn(workflow2.id, jm.workflows)
        self.assertEqual(workflow1, jm.workflows[workflow1.id])
        self.assertEqual(workflow2, jm.workflows[workflow2.id])



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

""" This file describe a new type of spreadsheet cell to embed
Matplotlib viewer into our spreadsheet
"""

from __future__ import division

from PyQt4 import QtCore, QtGui
import os

import matplotlib
import pylab
from matplotlib.backends.backend_qt4agg import FigureCanvasQTAgg, NavigationToolbar2QT
from matplotlib.backend_bases import NavigationToolbar2, FigureManagerBase

from vistrails.packages.spreadsheet.basic_widgets import SpreadsheetCell, SpreadsheetMode
from vistrails.packages.spreadsheet.spreadsheet_cell import QCellWidget, QCellToolBar

FigureCanvasQTAgg.DEBUG = True

################################################################################

class MplFigureToSpreadsheet(SpreadsheetMode):
    def compute_output(self, output_module, configuration=None):
        fig = output_module.get_input('value')
        self.display_and_wait(output_module, configuration,
                              MplFigureCellWidget, (fig,))

class MplFigureCell(SpreadsheetCell):
    """
    MplFigureCell is a spreadsheet cell for displaying Figure from
    Matplotlib

    """
    _input_ports = [("figure", "(MplFigure)")]

    def compute(self):
        """ compute() -> None        
        The class will take the figure manager and embed it into the spreadsheet
        
        """
        if self.has_input('figure'):
            fig = self.get_input('figure')
            self.displayAndWait(MplFigureCellWidget, (fig, ))

class MplFigureCellWidget(QCellWidget):
    """
    MplFigureCellWidget is the actual QWidget taking the FigureManager
    as a child for displaying figures

    """
    save_formats = ["Portable Document Format (*.pdf)",
                    "Portable Network Graphic (*.png)",
                    "PostScript (*.ps *.eps)",
                    "Raw images (*.raw *.rgba)",
                    "Scalable Vector Graphics (*.svg *.svgz)"]

    def __init__(self, parent=None):
        """ MplFigureCellWidget(parent: QWidget) -> MplFigureCellWidget
        Initialize the widget with its central layout
        
        """
        QCellWidget.__init__(self, parent)
        self.setFocusPolicy(QtCore.Qt.StrongFocus)
        centralLayout = QtGui.QVBoxLayout()
        self.setLayout(centralLayout)
        centralLayout.setMargin(0)
        centralLayout.setSpacing(0)
        self.canvas = None
        self.figure = None
        self.figManager = None
        self.toolBarType = MplFigureCellToolBar
        self.mplToolbar = None

    def updateContents(self, inputPorts):
        """ updateContents(inputPorts: tuple) -> None
        Update the widget contents based on the input data
        
        """
        (fig, ) = inputPorts
        if not self.figure or self.figure.number != fig.figInstance.number:
            if self.layout().count() > 0:
                self.layout().removeWidget(self.canvas)

            self.figure = fig.figInstance

            self.canvas = FigureCanvasQTAgg(self.figure)
            self.mplToolbar = MplNavigationToolbar(self.canvas, None)
            self.canvas.setSizePolicy(QtGui.QSizePolicy.Expanding,
                                      QtGui.QSizePolicy.Expanding)
            self.layout().addWidget(self.canvas)

    def keyPressEvent(self, event):
        print "KEY PRESS:",  event.key()
        self.canvas.keyPressEvent(event)

    def keyReleaseEvent(self, event):
        print "KEY RELEASE:", event.key()
        self.canvas.keyReleaseEvent(event)

    def deleteLater(self):
        """ deleteLater() -> None        
        Overriding PyQt deleteLater to free up resources
        
        """
        # Destroy the old one if possible
        if self.figure is not None:
            print "pylab:", pylab
            print "self.figure:", self.figure
            pylab.close(self.figure)

        QCellWidget.deleteLater(self)

    def grabWindowPixmap(self):
        """ grabWindowPixmap() -> QPixmap
        Widget special grabbing function
        
        """
        return QtGui.QPixmap.grabWidget(self.canvas)

    def dumpToFile(self, filename):
        previous_size = tuple(self.figure.get_size_inches())
        self.figure.set_size_inches(8.0,6.0)
        self.canvas.print_figure(filename)
        self.figure.set_size_inches(previous_size[0],previous_size[1])
        self.canvas.draw()
        
    def saveToPDF(self, filename):
        previous_size = tuple(self.figure.get_size_inches())
        self.figure.set_size_inches(8.0,6.0)
        self.canvas.print_figure(filename)
        self.figure.set_size_inches(previous_size[0],previous_size[1])
        self.canvas.draw()

class MplNavigationToolbar(NavigationToolbar2QT):
    # override a bunch of stuff here...
    def __init__(self, canvas, parent):
        self.canvas = canvas
        NavigationToolbar2.__init__(self, canvas)

    def _init_toolbar(self):
        self.adj_window = None

    def destroy(self):
        pass

    def pan(self, *args):
        NavigationToolbar2.pan(self, *args)
    
    def zoom(self, *args):
        NavigationToolbar2.zoom(self, *args)
        
    def set_message(self, s):
        pass

    def save_figure(self, *args):
        pass

    def set_history_buttons(self):
        pass

class MplFigureCellToolBar(QCellToolBar):
    def createToolBar(self):
        # can be copied from NavigationToolbar2QT... with checkable added
        toolitems = (
            ('Home', 'Reset original view', 'home.ppm', 'home', False),
            ('Back', 'Back to  previous view','back.ppm', 'back', False),
            ('Forward', 'Forward to next view','forward.ppm', 'forward', 
             False),
            ('Pan', 'Pan axes with left mouse, zoom with right', 'move.ppm', 
             'pan', True),
            ('Zoom', 'Zoom to rectangle','zoom_to_rect.ppm', 'zoom', True),
            )
        icondir = os.path.join(matplotlib.rcParams[ 'datapath' ],'images')
        exclusive_actions = {}
        actions = {}
        for (text, tooltip_text, image_file, callback, checkable) in toolitems:
            icon = QtGui.QIcon(os.path.join(icondir, image_file))
            action = QtGui.QAction(icon, text, self)
            action.setStatusTip(tooltip_text)
            action.setCheckable(checkable)
            actions[text] = action
            if text == 'Pan' or text == 'Zoom':
                exclusive_actions[text] = action

        def get_callback(act, cb, txt):
            print "CONSTRUCTING CALLBACK!", act, cb, txt
            def callback():
                print "GOT CALLBACK!!"
                cellWidget = act.toolBar.getSnappedWidget()
                print "CALLING callback", cb
                getattr(cellWidget.mplToolbar, cb)()
                if txt in exclusive_actions:
                    for t, a in exclusive_actions.iteritems():
                        if txt != t:
                            a.setChecked(False)
            return callback


        for (text, tooltip_text, image_file, callback, checkable) in toolitems:
            action = actions[text]
            action.triggeredSlot = get_callback(action, callback, text)
            self.appendAction(action)



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

"""generated automatically by auto_dao.py"""

from __future__ import division

import copy

class DBPortSpec(object):

    vtType = 'portSpec'

    def __init__(self, id=None, name=None, type=None, spec=None):
        self.__db_id = id
        self.__db_name = name
        self.__db_type = type
        self.__db_spec = spec
        self.is_dirty = True
    
    def __copy__(self):
        cp = DBPortSpec()
        cp.db_id = self.db_id
        cp.is_dirty = self.is_dirty
        cp.db_name = self.db_name
        cp.is_dirty = self.is_dirty
        cp.db_type = self.db_type
        cp.is_dirty = self.is_dirty
        cp.db_spec = self.db_spec
        cp.is_dirty = self.is_dirty
        return cp

    def db_children(self, parent=(None,None), orphan=False):
        children = []
        children.append((self, parent[0], parent[1]))
        return children
    def __get_db_id(self):
        return self.__db_id
    def __set_db_id(self, id):
        self.__db_id = id
        self.is_dirty = True
    db_id = property(__get_db_id, __set_db_id)
    def db_add_id(self, id):
        self.__db_id = id
    def db_change_id(self, id):
        self.__db_id = id
    def db_delete_id(self, id):
        self.__db_id = None
    
    def __get_db_name(self):
        return self.__db_name
    def __set_db_name(self, name):
        self.__db_name = name
        self.is_dirty = True
    db_name = property(__get_db_name, __set_db_name)
    def db_add_name(self, name):
        self.__db_name = name
    def db_change_name(self, name):
        self.__db_name = name
    def db_delete_name(self, name):
        self.__db_name = None
    
    def __get_db_type(self):
        return self.__db_type
    def __set_db_type(self, type):
        self.__db_type = type
        self.is_dirty = True
    db_type = property(__get_db_type, __set_db_type)
    def db_add_type(self, type):
        self.__db_type = type
    def db_change_type(self, type):
        self.__db_type = type
    def db_delete_type(self, type):
        self.__db_type = None
    
    def __get_db_spec(self):
        return self.__db_spec
    def __set_db_spec(self, spec):
        self.__db_spec = spec
        self.is_dirty = True
    db_spec = property(__get_db_spec, __set_db_spec)
    def db_add_spec(self, spec):
        self.__db_spec = spec
    def db_change_spec(self, spec):
        self.__db_spec = spec
    def db_delete_spec(self, spec):
        self.__db_spec = None
    
    def getPrimaryKey(self):
        return self.__db_id

class DBModule(object):

    vtType = 'module'

    def __init__(self, id=None, cache=None, name=None, location=None, functions=None, annotations=None, portSpecs=None):
        self.__db_id = id
        self.__db_cache = cache
        self.__db_name = name
        self.__db_location = location
        if functions is None:
            self.__db_functions = []
        else:
            self.__db_functions = functions
        if annotations is None:
            self.__db_annotations = {}
        else:
            self.__db_annotations = annotations
        if portSpecs is None:
            self.__db_portSpecs = []
        else:
            self.__db_portSpecs = portSpecs
        self.is_dirty = True
    
    def __copy__(self):
        cp = DBModule()
        cp.db_id = self.db_id
        cp.is_dirty = self.is_dirty
        cp.db_cache = self.db_cache
        cp.is_dirty = self.is_dirty
        cp.db_name = self.db_name
        cp.is_dirty = self.is_dirty
        cp.db_location = self.db_location
        cp.is_dirty = self.is_dirty
        if self.db_functions is None:
            cp.db_functions = None
        else:
            cp.db_functions = [copy.copy(v) for v in self.db_functions]
        cp.is_dirty = self.is_dirty
        if self.db_annotations is None:
            cp.db_annotations = None
        else:
            cp.db_annotations = dict([(k,copy.copy(v)) for (k,v) in self.db_annotations.iteritems()])
        cp.is_dirty = self.is_dirty
        if self.db_portSpecs is None:
            cp.db_portSpecs = None
        else:
            cp.db_portSpecs = [copy.copy(v) for v in self.db_portSpecs]
        cp.is_dirty = self.is_dirty
        return cp

    def db_children(self, parent=(None,None), orphan=False):
        children = []
        children.extend(self.db_location.db_children((self.vtType, self.db_id), orphan))
        if orphan:
            self.db_location = None
        for child in self.db_functions:
            children.extend(child.db_children((self.vtType, self.db_id), orphan))
        if orphan:
            self.db_functions = []
        for child in self.db_annotations.itervalues():
            children.extend(child.db_children((self.vtType, self.db_id), orphan))
        if orphan:
            self.db_annotations = {}
        for child in self.db_portSpecs:
            children.extend(child.db_children((self.vtType, self.db_id), orphan))
        if orphan:
            self.db_portSpecs = []
        children.append((self, parent[0], parent[1]))
        return children
    def __get_db_id(self):
        return self.__db_id
    def __set_db_id(self, id):
        self.__db_id = id
        self.is_dirty = True
    db_id = property(__get_db_id, __set_db_id)
    def db_add_id(self, id):
        self.__db_id = id
    def db_change_id(self, id):
        self.__db_id = id
    def db_delete_id(self, id):
        self.__db_id = None
    
    def __get_db_cache(self):
        return self.__db_cache
    def __set_db_cache(self, cache):
        self.__db_cache = cache
        self.is_dirty = True
    db_cache = property(__get_db_cache, __set_db_cache)
    def db_add_cache(self, cache):
        self.__db_cache = cache
    def db_change_cache(self, cache):
        self.__db_cache = cache
    def db_delete_cache(self, cache):
        self.__db_cache = None
    
    def __get_db_name(self):
        return self.__db_name
    def __set_db_name(self, name):
        self.__db_name = name
        self.is_dirty = True
    db_name = property(__get_db_name, __set_db_name)
    def db_add_name(self, name):
        self.__db_name = name
    def db_change_name(self, name):
        self.__db_name = name
    def db_delete_name(self, name):
        self.__db_name = None
    
    def __get_db_location(self):
        return self.__db_location
    def __set_db_location(self, location):
        self.__db_location = location
        self.is_dirty = True
    db_location = property(__get_db_location, __set_db_location)
    def db_add_location(self, location):
        self.__db_location = location
    def db_change_location(self, location):
        self.__db_location = location
    def db_delete_location(self, location):
        self.__db_location = None
    
    def __get_db_functions(self):
        return self.__db_functions
    def __set_db_functions(self, functions):
        self.__db_functions = functions
        self.is_dirty = True
    db_functions = property(__get_db_functions, __set_db_functions)
    def db_get_functions(self):
        return self.__db_functions
    def db_add_function(self, function):
        self.is_dirty = True
        self.__db_functions.append(function)
    def db_change_function(self, function):
        self.is_dirty = True
        found = False
        for i in xrange(len(self.__db_functions)):
            if self.__db_functions[i].db_id == function.db_id:
                self.__db_functions[i] = function
                found = True
                break
        if not found:
            self.__db_functions.append(function)
    def db_delete_function(self, function):
        self.is_dirty = True
        for i in xrange(len(self.__db_functions)):
            if self.__db_functions[i].db_id == function.db_id:
                del self.__db_functions[i]
                break
    def db_get_function(self, key):
        for i in xrange(len(self.__db_functions)):
            if self.__db_functions[i].db_id == key:
                return self.__db_functions[i]
        return None
    
    def __get_db_annotations(self):
        return self.__db_annotations
    def __set_db_annotations(self, annotations):
        self.__db_annotations = annotations
        self.is_dirty = True
    db_annotations = property(__get_db_annotations, __set_db_annotations)
    def db_get_annotations(self):
        return self.__db_annotations.values()
    def db_add_annotation(self, annotation):
        self.is_dirty = True
        self.__db_annotations[annotation.db_id] = annotation
    def db_change_annotation(self, annotation):
        self.is_dirty = True
        self.__db_annotations[annotation.db_id] = annotation
    def db_delete_annotation(self, annotation):
        self.is_dirty = True
        del self.__db_annotations[annotation.db_id]
    def db_get_annotation(self, key):
        if self.__db_annotations.has_key(key):
            return self.__db_annotations[key]
        return None
    
    def __get_db_portSpecs(self):
        return self.__db_portSpecs
    def __set_db_portSpecs(self, portSpecs):
        self.__db_portSpecs = portSpecs
        self.is_dirty = True
    db_portSpecs = property(__get_db_portSpecs, __set_db_portSpecs)
    def db_get_portSpecs(self):
        return self.__db_portSpecs
    def db_add_portSpec(self, portSpec):
        self.is_dirty = True
        self.__db_portSpecs.append(portSpec)
    def db_change_portSpec(self, portSpec):
        self.is_dirty = True
        found = False
        for i in xrange(len(self.__db_portSpecs)):
            if self.__db_portSpecs[i].db_id == portSpec.db_id:
                self.__db_portSpecs[i] = portSpec
                found = True
                break
        if not found:
            self.__db_portSpecs.append(portSpec)
    def db_delete_portSpec(self, portSpec):
        self.is_dirty = True
        for i in xrange(len(self.__db_portSpecs)):
            if self.__db_portSpecs[i].db_id == portSpec.db_id:
                del self.__db_portSpecs[i]
                break
    def db_get_portSpec(self, key):
        for i in xrange(len(self.__db_portSpecs)):
            if self.__db_portSpecs[i].db_id == key:
                return self.__db_portSpecs[i]
        return None
    
    def getPrimaryKey(self):
        return self.__db_id

class DBSession(object):

    vtType = 'session'

    def __init__(self, id=None, user=None, ip=None, visVersion=None, tsStart=None, tsEnd=None, machineId=None, wfExecs=None):
        self.__db_id = id
        self.__db_user = user
        self.__db_ip = ip
        self.__db_visVersion = visVersion
        self.__db_tsStart = tsStart
        self.__db_tsEnd = tsEnd
        self.__db_machineId = machineId
        if wfExecs is None:
            self.__db_wfExecs = {}
        else:
            self.__db_wfExecs = wfExecs
        self.is_dirty = True
    
    def __copy__(self):
        cp = DBSession()
        cp.db_id = self.db_id
        cp.is_dirty = self.is_dirty
        cp.db_user = self.db_user
        cp.is_dirty = self.is_dirty
        cp.db_ip = self.db_ip
        cp.is_dirty = self.is_dirty
        cp.db_visVersion = self.db_visVersion
        cp.is_dirty = self.is_dirty
        cp.db_tsStart = self.db_tsStart
        cp.is_dirty = self.is_dirty
        cp.db_tsEnd = self.db_tsEnd
        cp.is_dirty = self.is_dirty
        cp.db_machineId = self.db_machineId
        cp.is_dirty = self.is_dirty
        if self.db_wfExecs is None:
            cp.db_wfExecs = None
        else:
            cp.db_wfExecs = dict([(k,copy.copy(v)) for (k,v) in self.db_wfExecs.iteritems()])
        cp.is_dirty = self.is_dirty
        return cp

    def db_children(self, parent=(None,None), orphan=False):
        children = []
        for child in self.db_wfExecs.itervalues():
            children.extend(child.db_children((self.vtType, self.db_id), orphan))
        if orphan:
            self.db_wfExecs = {}
        children.extend(self.db_log.db_children((self.vtType, self.db_id), orphan))
        if orphan:
            self.db_log = None
        children.append((self, parent[0], parent[1]))
        return children
    def __get_db_id(self):
        return self.__db_id
    def __set_db_id(self, id):
        self.__db_id = id
        self.is_dirty = True
    db_id = property(__get_db_id, __set_db_id)
    def db_add_id(self, id):
        self.__db_id = id
    def db_change_id(self, id):
        self.__db_id = id
    def db_delete_id(self, id):
        self.__db_id = None
    
    def __get_db_user(self):
        return self.__db_user
    def __set_db_user(self, user):
        self.__db_user = user
        self.is_dirty = True
    db_user = property(__get_db_user, __set_db_user)
    def db_add_user(self, user):
        self.__db_user = user
    def db_change_user(self, user):
        self.__db_user = user
    def db_delete_user(self, user):
        self.__db_user = None
    
    def __get_db_ip(self):
        return self.__db_ip
    def __set_db_ip(self, ip):
        self.__db_ip = ip
        self.is_dirty = True
    db_ip = property(__get_db_ip, __set_db_ip)
    def db_add_ip(self, ip):
        self.__db_ip = ip
    def db_change_ip(self, ip):
        self.__db_ip = ip
    def db_delete_ip(self, ip):
        self.__db_ip = None
    
    def __get_db_visVersion(self):
        return self.__db_visVersion
    def __set_db_visVersion(self, visVersion):
        self.__db_visVersion = visVersion
        self.is_dirty = True
    db_visVersion = property(__get_db_visVersion, __set_db_visVersion)
    def db_add_visVersion(self, visVersion):
        self.__db_visVersion = visVersion
    def db_change_visVersion(self, visVersion):
        self.__db_visVersion = visVersion
    def db_delete_visVersion(self, visVersion):
        self.__db_visVersion = None
    
    def __get_db_tsStart(self):
        return self.__db_tsStart
    def __set_db_tsStart(self, tsStart):
        self.__db_tsStart = tsStart
        self.is_dirty = True
    db_tsStart = property(__get_db_tsStart, __set_db_tsStart)
    def db_add_tsStart(self, tsStart):
        self.__db_tsStart = tsStart
    def db_change_tsStart(self, tsStart):
        self.__db_tsStart = tsStart
    def db_delete_tsStart(self, tsStart):
        self.__db_tsStart = None
    
    def __get_db_tsEnd(self):
        return self.__db_tsEnd
    def __set_db_tsEnd(self, tsEnd):
        self.__db_tsEnd = tsEnd
        self.is_dirty = True
    db_tsEnd = property(__get_db_tsEnd, __set_db_tsEnd)
    def db_add_tsEnd(self, tsEnd):
        self.__db_tsEnd = tsEnd
    def db_change_tsEnd(self, tsEnd):
        self.__db_tsEnd = tsEnd
    def db_delete_tsEnd(self, tsEnd):
        self.__db_tsEnd = None
    
    def __get_db_machineId(self):
        return self.__db_machineId
    def __set_db_machineId(self, machineId):
        self.__db_machineId = machineId
        self.is_dirty = True
    db_machineId = property(__get_db_machineId, __set_db_machineId)
    def db_add_machineId(self, machineId):
        self.__db_machineId = machineId
    def db_change_machineId(self, machineId):
        self.__db_machineId = machineId
    def db_delete_machineId(self, machineId):
        self.__db_machineId = None
    
    def __get_db_wfExecs(self):
        return self.__db_wfExecs
    def __set_db_wfExecs(self, wfExecs):
        self.__db_wfExecs = wfExecs
        self.is_dirty = True
    db_wfExecs = property(__get_db_wfExecs, __set_db_wfExecs)
    def db_get_wfExecs(self):
        return self.__db_wfExecs.values()
    def db_add_wfExec(self, wfExec):
        self.is_dirty = True
        self.__db_wfExecs[wfExec.db_id] = wfExec
    def db_change_wfExec(self, wfExec):
        self.is_dirty = True
        self.__db_wfExecs[wfExec.db_id] = wfExec
    def db_delete_wfExec(self, wfExec):
        self.is_dirty = True
        del self.__db_wfExecs[wfExec.db_id]
    def db_get_wfExec(self, key):
        if self.__db_wfExecs.has_key(key):
            return self.__db_wfExecs[key]
        return None
    
    def getPrimaryKey(self):
        return self.__db_id

class DBPort(object):

    vtType = 'port'

    def __init__(self, id=None, type=None, moduleId=None, moduleName=None, sig=None):
        self.__db_id = id
        self.__db_type = type
        self.__db_moduleId = moduleId
        self.__db_moduleName = moduleName
        self.__db_sig = sig
        self.is_dirty = True
    
    def __copy__(self):
        cp = DBPort()
        cp.db_id = self.db_id
        cp.is_dirty = self.is_dirty
        cp.db_type = self.db_type
        cp.is_dirty = self.is_dirty
        cp.db_moduleId = self.db_moduleId
        cp.is_dirty = self.is_dirty
        cp.db_moduleName = self.db_moduleName
        cp.is_dirty = self.is_dirty
        cp.db_sig = self.db_sig
        cp.is_dirty = self.is_dirty
        return cp

    def db_children(self, parent=(None,None), orphan=False):
        children = []
        children.append((self, parent[0], parent[1]))
        return children
    def __get_db_id(self):
        return self.__db_id
    def __set_db_id(self, id):
        self.__db_id = id
        self.is_dirty = True
    db_id = property(__get_db_id, __set_db_id)
    def db_add_id(self, id):
        self.__db_id = id
    def db_change_id(self, id):
        self.__db_id = id
    def db_delete_id(self, id):
        self.__db_id = None
    
    def __get_db_type(self):
        return self.__db_type
    def __set_db_type(self, type):
        self.__db_type = type
        self.is_dirty = True
    db_type = property(__get_db_type, __set_db_type)
    def db_add_type(self, type):
        self.__db_type = type
    def db_change_type(self, type):
        self.__db_type = type
    def db_delete_type(self, type):
        self.__db_type = None
    
    def __get_db_moduleId(self):
        return self.__db_moduleId
    def __set_db_moduleId(self, moduleId):
        self.__db_moduleId = moduleId
        self.is_dirty = True
    db_moduleId = property(__get_db_moduleId, __set_db_moduleId)
    def db_add_moduleId(self, moduleId):
        self.__db_moduleId = moduleId
    def db_change_moduleId(self, moduleId):
        self.__db_moduleId = moduleId
    def db_delete_moduleId(self, moduleId):
        self.__db_moduleId = None
    
    def __get_db_moduleName(self):
        return self.__db_moduleName
    def __set_db_moduleName(self, moduleName):
        self.__db_moduleName = moduleName
        self.is_dirty = True
    db_moduleName = property(__get_db_moduleName, __set_db_moduleName)
    def db_add_moduleName(self, moduleName):
        self.__db_moduleName = moduleName
    def db_change_moduleName(self, moduleName):
        self.__db_moduleName = moduleName
    def db_delete_moduleName(self, moduleName):
        self.__db_moduleName = None
    
    def __get_db_sig(self):
        return self.__db_sig
    def __set_db_sig(self, sig):
        self.__db_sig = sig
        self.is_dirty = True
    db_sig = property(__get_db_sig, __set_db_sig)
    def db_add_sig(self, sig):
        self.__db_sig = sig
    def db_change_sig(self, sig):
        self.__db_sig = sig
    def db_delete_sig(self, sig):
        self.__db_sig = None
    
    def getPrimaryKey(self):
        return self.__db_id

class DBLog(object):

    vtType = 'log'

    def __init__(self, id=None, sessions=None, machines=None):
        self.__db_id = id
        if sessions is None:
            self.__db_sessions = {}
        else:
            self.__db_sessions = sessions
        if machines is None:
            self.__db_machines = {}
        else:
            self.__db_machines = machines
        self.is_dirty = True
    
    def __copy__(self):
        cp = DBLog()
        cp.db_id = self.db_id
        cp.is_dirty = self.is_dirty
        if self.db_sessions is None:
            cp.db_sessions = None
        else:
            cp.db_sessions = dict([(k,copy.copy(v)) for (k,v) in self.db_sessions.iteritems()])
        cp.is_dirty = self.is_dirty
        if self.db_machines is None:
            cp.db_machines = None
        else:
            cp.db_machines = dict([(k,copy.copy(v)) for (k,v) in self.db_machines.iteritems()])
        cp.is_dirty = self.is_dirty
        return cp

    def db_children(self, parent=(None,None), orphan=False):
        children = []
        for child in self.db_sessions.itervalues():
            children.extend(child.db_children((self.vtType, self.db_id), orphan))
        if orphan:
            self.db_sessions = {}
        for child in self.db_machines.itervalues():
            children.extend(child.db_children((self.vtType, self.db_id), orphan))
        if orphan:
            self.db_machines = {}
        children.append((self, parent[0], parent[1]))
        return children
    def __get_db_id(self):
        return self.__db_id
    def __set_db_id(self, id):
        self.__db_id = id
        self.is_dirty = True
    db_id = property(__get_db_id, __set_db_id)
    def db_add_id(self, id):
        self.__db_id = id
    def db_change_id(self, id):
        self.__db_id = id
    def db_delete_id(self, id):
        self.__db_id = None
    
    def __get_db_sessions(self):
        return self.__db_sessions
    def __set_db_sessions(self, sessions):
        self.__db_sessions = sessions
        self.is_dirty = True
    db_sessions = property(__get_db_sessions, __set_db_sessions)
    def db_get_sessions(self):
        return self.__db_sessions.values()
    def db_add_session(self, session):
        self.is_dirty = True
        self.__db_sessions[session.db_id] = session
    def db_change_session(self, session):
        self.is_dirty = True
        self.__db_sessions[session.db_id] = session
    def db_delete_session(self, session):
        self.is_dirty = True
        del self.__db_sessions[session.db_id]
    def db_get_session(self, key):
        if self.__db_sessions.has_key(key):
            return self.__db_sessions[key]
        return None
    
    def __get_db_machines(self):
        return self.__db_machines
    def __set_db_machines(self, machines):
        self.__db_machines = machines
        self.is_dirty = True
    db_machines = property(__get_db_machines, __set_db_machines)
    def db_get_machines(self):
        return self.__db_machines.values()
    def db_add_machine(self, machine):
        self.is_dirty = True
        self.__db_machines[machine.db_id] = machine
    def db_change_machine(self, machine):
        self.is_dirty = True
        self.__db_machines[machine.db_id] = machine
    def db_delete_machine(self, machine):
        self.is_dirty = True
        del self.__db_machines[machine.db_id]
    def db_get_machine(self, key):
        if self.__db_machines.has_key(key):
            return self.__db_machines[key]
        return None
    
    def getPrimaryKey(self):
        return self.__db_id

class DBMachine(object):

    vtType = 'machine'

    def __init__(self, id=None, name=None, os=None, architecture=None, processor=None, ram=None):
        self.__db_id = id
        self.__db_name = name
        self.__db_os = os
        self.__db_architecture = architecture
        self.__db_processor = processor
        self.__db_ram = ram
        self.is_dirty = True
    
    def __copy__(self):
        cp = DBMachine()
        cp.db_id = self.db_id
        cp.is_dirty = self.is_dirty
        cp.db_name = self.db_name
        cp.is_dirty = self.is_dirty
        cp.db_os = self.db_os
        cp.is_dirty = self.is_dirty
        cp.db_architecture = self.db_architecture
        cp.is_dirty = self.is_dirty
        cp.db_processor = self.db_processor
        cp.is_dirty = self.is_dirty
        cp.db_ram = self.db_ram
        cp.is_dirty = self.is_dirty
        return cp

    def db_children(self, parent=(None,None), orphan=False):
        children = []
        children.extend(self.db_log.db_children((self.vtType, self.db_id), orphan))
        if orphan:
            self.db_log = None
        children.append((self, parent[0], parent[1]))
        return children
    def __get_db_id(self):
        return self.__db_id
    def __set_db_id(self, id):
        self.__db_id = id
        self.is_dirty = True
    db_id = property(__get_db_id, __set_db_id)
    def db_add_id(self, id):
        self.__db_id = id
    def db_change_id(self, id):
        self.__db_id = id
    def db_delete_id(self, id):
        self.__db_id = None
    
    def __get_db_name(self):
        return self.__db_name
    def __set_db_name(self, name):
        self.__db_name = name
        self.is_dirty = True
    db_name = property(__get_db_name, __set_db_name)
    def db_add_name(self, name):
        self.__db_name = name
    def db_change_name(self, name):
        self.__db_name = name
    def db_delete_name(self, name):
        self.__db_name = None
    
    def __get_db_os(self):
        return self.__db_os
    def __set_db_os(self, os):
        self.__db_os = os
        self.is_dirty = True
    db_os = property(__get_db_os, __set_db_os)
    def db_add_os(self, os):
        self.__db_os = os
    def db_change_os(self, os):
        self.__db_os = os
    def db_delete_os(self, os):
        self.__db_os = None
    
    def __get_db_architecture(self):
        return self.__db_architecture
    def __set_db_architecture(self, architecture):
        self.__db_architecture = architecture
        self.is_dirty = True
    db_architecture = property(__get_db_architecture, __set_db_architecture)
    def db_add_architecture(self, architecture):
        self.__db_architecture = architecture
    def db_change_architecture(self, architecture):
        self.__db_architecture = architecture
    def db_delete_architecture(self, architecture):
        self.__db_architecture = None
    
    def __get_db_processor(self):
        return self.__db_processor
    def __set_db_processor(self, processor):
        self.__db_processor = processor
        self.is_dirty = True
    db_processor = property(__get_db_processor, __set_db_processor)
    def db_add_processor(self, processor):
        self.__db_processor = processor
    def db_change_processor(self, processor):
        self.__db_processor = processor
    def db_delete_processor(self, processor):
        self.__db_processor = None
    
    def __get_db_ram(self):
        return self.__db_ram
    def __set_db_ram(self, ram):
        self.__db_ram = ram
        self.is_dirty = True
    db_ram = property(__get_db_ram, __set_db_ram)
    def db_add_ram(self, ram):
        self.__db_ram = ram
    def db_change_ram(self, ram):
        self.__db_ram = ram
    def db_delete_ram(self, ram):
        self.__db_ram = None
    
    def getPrimaryKey(self):
        return self.__db_id

class DBAdd(object):

    vtType = 'add'

    def __init__(self, id=None, what=None, objectId=None, parentObjId=None, parentObjType=None, data=None):
        self.__db_id = id
        self.__db_what = what
        self.__db_objectId = objectId
        self.__db_parentObjId = parentObjId
        self.__db_parentObjType = parentObjType
        self.__db_data = data
        self.is_dirty = True
    
    def __copy__(self):
        cp = DBAdd()
        cp.db_id = self.db_id
        cp.is_dirty = self.is_dirty
        cp.db_what = self.db_what
        cp.is_dirty = self.is_dirty
        cp.db_objectId = self.db_objectId
        cp.is_dirty = self.is_dirty
        cp.db_parentObjId = self.db_parentObjId
        cp.is_dirty = self.is_dirty
        cp.db_parentObjType = self.db_parentObjType
        cp.is_dirty = self.is_dirty
        cp.db_data = self.db_data
        cp.is_dirty = self.is_dirty
        return cp

    def db_children(self, parent=(None,None), orphan=False):
        children = []
        children.extend(self.db_action.db_children((self.vtType, self.db_id), orphan))
        if orphan:
            self.db_action = None
        children.append((self, parent[0], parent[1]))
        return children
    def __get_db_id(self):
        return self.__db_id
    def __set_db_id(self, id):
        self.__db_id = id
        self.is_dirty = True
    db_id = property(__get_db_id, __set_db_id)
    def db_add_id(self, id):
        self.__db_id = id
    def db_change_id(self, id):
        self.__db_id = id
    def db_delete_id(self, id):
        self.__db_id = None
    
    def __get_db_what(self):
        return self.__db_what
    def __set_db_what(self, what):
        self.__db_what = what
        self.is_dirty = True
    db_what = property(__get_db_what, __set_db_what)
    def db_add_what(self, what):
        self.__db_what = what
    def db_change_what(self, what):
        self.__db_what = what
    def db_delete_what(self, what):
        self.__db_what = None
    
    def __get_db_objectId(self):
        return self.__db_objectId
    def __set_db_objectId(self, objectId):
        self.__db_objectId = objectId
        self.is_dirty = True
    db_objectId = property(__get_db_objectId, __set_db_objectId)
    def db_add_objectId(self, objectId):
        self.__db_objectId = objectId
    def db_change_objectId(self, objectId):
        self.__db_objectId = objectId
    def db_delete_objectId(self, objectId):
        self.__db_objectId = None
    
    def __get_db_parentObjId(self):
        return self.__db_parentObjId
    def __set_db_parentObjId(self, parentObjId):
        self.__db_parentObjId = parentObjId
        self.is_dirty = True
    db_parentObjId = property(__get_db_parentObjId, __set_db_parentObjId)
    def db_add_parentObjId(self, parentObjId):
        self.__db_parentObjId = parentObjId
    def db_change_parentObjId(self, parentObjId):
        self.__db_parentObjId = parentObjId
    def db_delete_parentObjId(self, parentObjId):
        self.__db_parentObjId = None
    
    def __get_db_parentObjType(self):
        return self.__db_parentObjType
    def __set_db_parentObjType(self, parentObjType):
        self.__db_parentObjType = parentObjType
        self.is_dirty = True
    db_parentObjType = property(__get_db_parentObjType, __set_db_parentObjType)
    def db_add_parentObjType(self, parentObjType):
        self.__db_parentObjType = parentObjType
    def db_change_parentObjType(self, parentObjType):
        self.__db_parentObjType = parentObjType
    def db_delete_parentObjType(self, parentObjType):
        self.__db_parentObjType = None
    
    def __get_db_data(self):
        return self.__db_data
    def __set_db_data(self, data):
        self.__db_data = data
        self.is_dirty = True
    db_data = property(__get_db_data, __set_db_data)
    def db_add_data(self, data):
        self.__db_data = data
    def db_change_data(self, data):
        self.__db_data = data
    def db_delete_data(self, data):
        self.__db_data = None
    
    def getPrimaryKey(self):
        return self.__db_id

class DBOther(object):

    vtType = 'other'

    def __init__(self, id=None, key=None, value=None):
        self.__db_id = id
        self.__db_key = key
        self.__db_value = value
        self.is_dirty = True
    
    def __copy__(self):
        cp = DBOther()
        cp.db_id = self.db_id
        cp.is_dirty = self.is_dirty
        cp.db_key = self.db_key
        cp.is_dirty = self.is_dirty
        cp.db_value = self.db_value
        cp.is_dirty = self.is_dirty
        return cp

    def db_children(self, parent=(None,None), orphan=False):
        children = []
        children.append((self, parent[0], parent[1]))
        return children
    def __get_db_id(self):
        return self.__db_id
    def __set_db_id(self, id):
        self.__db_id = id
        self.is_dirty = True
    db_id = property(__get_db_id, __set_db_id)
    def db_add_id(self, id):
        self.__db_id = id
    def db_change_id(self, id):
        self.__db_id = id
    def db_delete_id(self, id):
        self.__db_id = None
    
    def __get_db_key(self):
        return self.__db_key
    def __set_db_key(self, key):
        self.__db_key = key
        self.is_dirty = True
    db_key = property(__get_db_key, __set_db_key)
    def db_add_key(self, key):
        self.__db_key = key
    def db_change_key(self, key):
        self.__db_key = key
    def db_delete_key(self, key):
        self.__db_key = None
    
    def __get_db_value(self):
        return self.__db_value
    def __set_db_value(self, value):
        self.__db_value = value
        self.is_dirty = True
    db_value = property(__get_db_value, __set_db_value)
    def db_add_value(self, value):
        self.__db_value = value
    def db_change_value(self, value):
        self.__db_value = value
    def db_delete_value(self, value):
        self.__db_value = None
    
    def getPrimaryKey(self):
        return self.__db_id

class DBLocation(object):

    vtType = 'location'

    def __init__(self, id=None, x=None, y=None):
        self.__db_id = id
        self.__db_x = x
        self.__db_y = y
        self.is_dirty = True
    
    def __copy__(self):
        cp = DBLocation()
        cp.db_id = self.db_id
        cp.is_dirty = self.is_dirty
        cp.db_x = self.db_x
        cp.is_dirty = self.is_dirty
        cp.db_y = self.db_y
        cp.is_dirty = self.is_dirty
        return cp

    def db_children(self, parent=(None,None), orphan=False):
        children = []
        children.append((self, parent[0], parent[1]))
        return children
    def __get_db_id(self):
        return self.__db_id
    def __set_db_id(self, id):
        self.__db_id = id
        self.is_dirty = True
    db_id = property(__get_db_id, __set_db_id)
    def db_add_id(self, id):
        self.__db_id = id
    def db_change_id(self, id):
        self.__db_id = id
    def db_delete_id(self, id):
        self.__db_id = None
    
    def __get_db_x(self):
        return self.__db_x
    def __set_db_x(self, x):
        self.__db_x = x
        self.is_dirty = True
    db_x = property(__get_db_x, __set_db_x)
    def db_add_x(self, x):
        self.__db_x = x
    def db_change_x(self, x):
        self.__db_x = x
    def db_delete_x(self, x):
        self.__db_x = None
    
    def __get_db_y(self):
        return self.__db_y
    def __set_db_y(self, y):
        self.__db_y = y
        self.is_dirty = True
    db_y = property(__get_db_y, __set_db_y)
    def db_add_y(self, y):
        self.__db_y = y
    def db_change_y(self, y):
        self.__db_y = y
    def db_delete_y(self, y):
        self.__db_y = None
    
    def getPrimaryKey(self):
        return self.__db_id

class DBWfExec(object):

    vtType = 'wfExec'

    def __init__(self, id=None, tsStart=None, tsEnd=None, wfVersion=None, vistrailId=None, vistrailName=None, execRecs=None):
        self.__db_id = id
        self.__db_tsStart = tsStart
        self.__db_tsEnd = tsEnd
        self.__db_wfVersion = wfVersion
        self.__db_vistrailId = vistrailId
        self.__db_vistrailName = vistrailName
        if execRecs is None:
            self.__db_execRecs = {}
        else:
            self.__db_execRecs = execRecs
        self.is_dirty = True
    
    def __copy__(self):
        cp = DBWfExec()
        cp.db_id = self.db_id
        cp.is_dirty = self.is_dirty
        cp.db_tsStart = self.db_tsStart
        cp.is_dirty = self.is_dirty
        cp.db_tsEnd = self.db_tsEnd
        cp.is_dirty = self.is_dirty
        cp.db_wfVersion = self.db_wfVersion
        cp.is_dirty = self.is_dirty
        cp.db_vistrailId = self.db_vistrailId
        cp.is_dirty = self.is_dirty
        cp.db_vistrailName = self.db_vistrailName
        cp.is_dirty = self.is_dirty
        if self.db_execRecs is None:
            cp.db_execRecs = None
        else:
            cp.db_execRecs = dict([(k,copy.copy(v)) for (k,v) in self.db_execRecs.iteritems()])
        cp.is_dirty = self.is_dirty
        return cp

    def db_children(self, parent=(None,None), orphan=False):
        children = []
        children.extend(self.db_session.db_children((self.vtType, self.db_id), orphan))
        if orphan:
            self.db_session = None
        for child in self.db_execRecs.itervalues():
            children.extend(child.db_children((self.vtType, self.db_id), orphan))
        if orphan:
            self.db_execRecs = {}
        children.append((self, parent[0], parent[1]))
        return children
    def __get_db_id(self):
        return self.__db_id
    def __set_db_id(self, id):
        self.__db_id = id
        self.is_dirty = True
    db_id = property(__get_db_id, __set_db_id)
    def db_add_id(self, id):
        self.__db_id = id
    def db_change_id(self, id):
        self.__db_id = id
    def db_delete_id(self, id):
        self.__db_id = None
    
    def __get_db_tsStart(self):
        return self.__db_tsStart
    def __set_db_tsStart(self, tsStart):
        self.__db_tsStart = tsStart
        self.is_dirty = True
    db_tsStart = property(__get_db_tsStart, __set_db_tsStart)
    def db_add_tsStart(self, tsStart):
        self.__db_tsStart = tsStart
    def db_change_tsStart(self, tsStart):
        self.__db_tsStart = tsStart
    def db_delete_tsStart(self, tsStart):
        self.__db_tsStart = None
    
    def __get_db_tsEnd(self):
        return self.__db_tsEnd
    def __set_db_tsEnd(self, tsEnd):
        self.__db_tsEnd = tsEnd
        self.is_dirty = True
    db_tsEnd = property(__get_db_tsEnd, __set_db_tsEnd)
    def db_add_tsEnd(self, tsEnd):
        self.__db_tsEnd = tsEnd
    def db_change_tsEnd(self, tsEnd):
        self.__db_tsEnd = tsEnd
    def db_delete_tsEnd(self, tsEnd):
        self.__db_tsEnd = None
    
    def __get_db_wfVersion(self):
        return self.__db_wfVersion
    def __set_db_wfVersion(self, wfVersion):
        self.__db_wfVersion = wfVersion
        self.is_dirty = True
    db_wfVersion = property(__get_db_wfVersion, __set_db_wfVersion)
    def db_add_wfVersion(self, wfVersion):
        self.__db_wfVersion = wfVersion
    def db_change_wfVersion(self, wfVersion):
        self.__db_wfVersion = wfVersion
    def db_delete_wfVersion(self, wfVersion):
        self.__db_wfVersion = None
    
    def __get_db_vistrailId(self):
        return self.__db_vistrailId
    def __set_db_vistrailId(self, vistrailId):
        self.__db_vistrailId = vistrailId
        self.is_dirty = True
    db_vistrailId = property(__get_db_vistrailId, __set_db_vistrailId)
    def db_add_vistrailId(self, vistrailId):
        self.__db_vistrailId = vistrailId
    def db_change_vistrailId(self, vistrailId):
        self.__db_vistrailId = vistrailId
    def db_delete_vistrailId(self, vistrailId):
        self.__db_vistrailId = None
    
    def __get_db_vistrailName(self):
        return self.__db_vistrailName
    def __set_db_vistrailName(self, vistrailName):
        self.__db_vistrailName = vistrailName
        self.is_dirty = True
    db_vistrailName = property(__get_db_vistrailName, __set_db_vistrailName)
    def db_add_vistrailName(self, vistrailName):
        self.__db_vistrailName = vistrailName
    def db_change_vistrailName(self, vistrailName):
        self.__db_vistrailName = vistrailName
    def db_delete_vistrailName(self, vistrailName):
        self.__db_vistrailName = None
    
    def __get_db_execRecs(self):
        return self.__db_execRecs
    def __set_db_execRecs(self, execRecs):
        self.__db_execRecs = execRecs
        self.is_dirty = True
    db_execRecs = property(__get_db_execRecs, __set_db_execRecs)
    def db_get_execRecs(self):
        return self.__db_execRecs.values()
    def db_add_execRec(self, execRec):
        self.is_dirty = True
        self.__db_execRecs[execRec.db_id] = execRec
    def db_change_execRec(self, execRec):
        self.is_dirty = True
        self.__db_execRecs[execRec.db_id] = execRec
    def db_delete_execRec(self, execRec):
        self.is_dirty = True
        del self.__db_execRecs[execRec.db_id]
    def db_get_execRec(self, key):
        if self.__db_execRecs.has_key(key):
            return self.__db_execRecs[key]
        return None
    
    def getPrimaryKey(self):
        return self.__db_id

class DBParameter(object):

    vtType = 'parameter'

    def __init__(self, id=None, pos=None, name=None, type=None, val=None, alias=None):
        self.__db_id = id
        self.__db_pos = pos
        self.__db_name = name
        self.__db_type = type
        self.__db_val = val
        self.__db_alias = alias
        self.is_dirty = True
    
    def __copy__(self):
        cp = DBParameter()
        cp.db_id = self.db_id
        cp.is_dirty = self.is_dirty
        cp.db_pos = self.db_pos
        cp.is_dirty = self.is_dirty
        cp.db_name = self.db_name
        cp.is_dirty = self.is_dirty
        cp.db_type = self.db_type
        cp.is_dirty = self.is_dirty
        cp.db_val = self.db_val
        cp.is_dirty = self.is_dirty
        cp.db_alias = self.db_alias
        cp.is_dirty = self.is_dirty
        return cp

    def db_children(self, parent=(None,None), orphan=False):
        children = []
        children.append((self, parent[0], parent[1]))
        return children
    def __get_db_id(self):
        return self.__db_id
    def __set_db_id(self, id):
        self.__db_id = id
        self.is_dirty = True
    db_id = property(__get_db_id, __set_db_id)
    def db_add_id(self, id):
        self.__db_id = id
    def db_change_id(self, id):
        self.__db_id = id
    def db_delete_id(self, id):
        self.__db_id = None
    
    def __get_db_pos(self):
        return self.__db_pos
    def __set_db_pos(self, pos):
        self.__db_pos = pos
        self.is_dirty = True
    db_pos = property(__get_db_pos, __set_db_pos)
    def db_add_pos(self, pos):
        self.__db_pos = pos
    def db_change_pos(self, pos):
        self.__db_pos = pos
    def db_delete_pos(self, pos):
        self.__db_pos = None
    
    def __get_db_name(self):
        return self.__db_name
    def __set_db_name(self, name):
        self.__db_name = name
        self.is_dirty = True
    db_name = property(__get_db_name, __set_db_name)
    def db_add_name(self, name):
        self.__db_name = name
    def db_change_name(self, name):
        self.__db_name = name
    def db_delete_name(self, name):
        self.__db_name = None
    
    def __get_db_type(self):
        return self.__db_type
    def __set_db_type(self, type):
        self.__db_type = type
        self.is_dirty = True
    db_type = property(__get_db_type, __set_db_type)
    def db_add_type(self, type):
        self.__db_type = type
    def db_change_type(self, type):
        self.__db_type = type
    def db_delete_type(self, type):
        self.__db_type = None
    
    def __get_db_val(self):
        return self.__db_val
    def __set_db_val(self, val):
        self.__db_val = val
        self.is_dirty = True
    db_val = property(__get_db_val, __set_db_val)
    def db_add_val(self, val):
        self.__db_val = val
    def db_change_val(self, val):
        self.__db_val = val
    def db_delete_val(self, val):
        self.__db_val = None
    
    def __get_db_alias(self):
        return self.__db_alias
    def __set_db_alias(self, alias):
        self.__db_alias = alias
        self.is_dirty = True
    db_alias = property(__get_db_alias, __set_db_alias)
    def db_add_alias(self, alias):
        self.__db_alias = alias
    def db_change_alias(self, alias):
        self.__db_alias = alias
    def db_delete_alias(self, alias):
        self.__db_alias = None
    
    def getPrimaryKey(self):
        return self.__db_id

class DBFunction(object):

    vtType = 'function'

    def __init__(self, id=None, pos=None, name=None, parameters=None):
        self.__db_id = id
        self.__db_pos = pos
        self.__db_name = name
        if parameters is None:
            self.__db_parameters = []
        else:
            self.__db_parameters = parameters
        self.is_dirty = True
    
    def __copy__(self):
        cp = DBFunction()
        cp.db_id = self.db_id
        cp.is_dirty = self.is_dirty
        cp.db_pos = self.db_pos
        cp.is_dirty = self.is_dirty
        cp.db_name = self.db_name
        cp.is_dirty = self.is_dirty
        if self.db_parameters is None:
            cp.db_parameters = None
        else:
            cp.db_parameters = [copy.copy(v) for v in self.db_parameters]
        cp.is_dirty = self.is_dirty
        return cp

    def db_children(self, parent=(None,None), orphan=False):
        children = []
        for child in self.db_parameters:
            children.extend(child.db_children((self.vtType, self.db_id), orphan))
        if orphan:
            self.db_parameters = []
        children.append((self, parent[0], parent[1]))
        return children
    def __get_db_id(self):
        return self.__db_id
    def __set_db_id(self, id):
        self.__db_id = id
        self.is_dirty = True
    db_id = property(__get_db_id, __set_db_id)
    def db_add_id(self, id):
        self.__db_id = id
    def db_change_id(self, id):
        self.__db_id = id
    def db_delete_id(self, id):
        self.__db_id = None
    
    def __get_db_pos(self):
        return self.__db_pos
    def __set_db_pos(self, pos):
        self.__db_pos = pos
        self.is_dirty = True
    db_pos = property(__get_db_pos, __set_db_pos)
    def db_add_pos(self, pos):
        self.__db_pos = pos
    def db_change_pos(self, pos):
        self.__db_pos = pos
    def db_delete_pos(self, pos):
        self.__db_pos = None
    
    def __get_db_name(self):
        return self.__db_name
    def __set_db_name(self, name):
        self.__db_name = name
        self.is_dirty = True
    db_name = property(__get_db_name, __set_db_name)
    def db_add_name(self, name):
        self.__db_name = name
    def db_change_name(self, name):
        self.__db_name = name
    def db_delete_name(self, name):
        self.__db_name = None
    
    def __get_db_parameters(self):
        return self.__db_parameters
    def __set_db_parameters(self, parameters):
        self.__db_parameters = parameters
        self.is_dirty = True
    db_parameters = property(__get_db_parameters, __set_db_parameters)
    def db_get_parameters(self):
        return self.__db_parameters
    def db_add_parameter(self, parameter):
        self.is_dirty = True
        self.__db_parameters.append(parameter)
    def db_change_parameter(self, parameter):
        self.is_dirty = True
        found = False
        for i in xrange(len(self.__db_parameters)):
            if self.__db_parameters[i].db_id == parameter.db_id:
                self.__db_parameters[i] = parameter
                found = True
                break
        if not found:
            self.__db_parameters.append(parameter)
    def db_delete_parameter(self, parameter):
        self.is_dirty = True
        for i in xrange(len(self.__db_parameters)):
            if self.__db_parameters[i].db_id == parameter.db_id:
                del self.__db_parameters[i]
                break
    def db_get_parameter(self, key):
        for i in xrange(len(self.__db_parameters)):
            if self.__db_parameters[i].db_id == key:
                return self.__db_parameters[i]
        return None
    
    def getPrimaryKey(self):
        return self.__db_id

class DBWorkflow(object):

    vtType = 'workflow'

    def __init__(self, id=None, name=None, modules=None, connections=None, annotations=None, others=None):
        self.__db_id = id
        self.__db_name = name
        if modules is None:
            self.__db_modules = {}
        else:
            self.__db_modules = modules
        if connections is None:
            self.__db_connections = {}
        else:
            self.__db_connections = connections
        if annotations is None:
            self.__db_annotations = []
        else:
            self.__db_annotations = annotations
        if others is None:
            self.__db_others = []
        else:
            self.__db_others = others
        self.is_dirty = True
    
    def __copy__(self):
        cp = DBWorkflow()
        cp.db_id = self.db_id
        cp.is_dirty = self.is_dirty
        cp.db_name = self.db_name
        cp.is_dirty = self.is_dirty
        if self.db_modules is None:
            cp.db_modules = None
        else:
            cp.db_modules = dict([(k,copy.copy(v)) for (k,v) in self.db_modules.iteritems()])
        cp.is_dirty = self.is_dirty
        if self.db_connections is None:
            cp.db_connections = None
        else:
            cp.db_connections = dict([(k,copy.copy(v)) for (k,v) in self.db_connections.iteritems()])
        cp.is_dirty = self.is_dirty
        if self.db_annotations is None:
            cp.db_annotations = None
        else:
            cp.db_annotations = [copy.copy(v) for v in self.db_annotations]
        cp.is_dirty = self.is_dirty
        if self.db_others is None:
            cp.db_others = None
        else:
            cp.db_others = [copy.copy(v) for v in self.db_others]
        cp.is_dirty = self.is_dirty
        return cp

    def db_children(self, parent=(None,None), orphan=False):
        children = []
        for child in self.db_modules.itervalues():
            children.extend(child.db_children((self.vtType, self.db_id), orphan))
        if orphan:
            self.db_modules = {}
        for child in self.db_connections.itervalues():
            children.extend(child.db_children((self.vtType, self.db_id), orphan))
        if orphan:
            self.db_connections = {}
        for child in self.db_annotations:
            children.extend(child.db_children((self.vtType, self.db_id), orphan))
        if orphan:
            self.db_annotations = []
        for child in self.db_others:
            children.extend(child.db_children((self.vtType, self.db_id), orphan))
        if orphan:
            self.db_others = []
        children.append((self, parent[0], parent[1]))
        return children
    def __get_db_id(self):
        return self.__db_id
    def __set_db_id(self, id):
        self.__db_id = id
        self.is_dirty = True
    db_id = property(__get_db_id, __set_db_id)
    def db_add_id(self, id):
        self.__db_id = id
    def db_change_id(self, id):
        self.__db_id = id
    def db_delete_id(self, id):
        self.__db_id = None
    
    def __get_db_name(self):
        return self.__db_name
    def __set_db_name(self, name):
        self.__db_name = name
        self.is_dirty = True
    db_name = property(__get_db_name, __set_db_name)
    def db_add_name(self, name):
        self.__db_name = name
    def db_change_name(self, name):
        self.__db_name = name
    def db_delete_name(self, name):
        self.__db_name = None
    
    def __get_db_modules(self):
        return self.__db_modules
    def __set_db_modules(self, modules):
        self.__db_modules = modules
        self.is_dirty = True
    db_modules = property(__get_db_modules, __set_db_modules)
    def db_get_modules(self):
        return self.__db_modules.values()
    def db_add_module(self, module):
        self.is_dirty = True
        self.__db_modules[module.db_id] = module
    def db_change_module(self, module):
        self.is_dirty = True
        self.__db_modules[module.db_id] = module
    def db_delete_module(self, module):
        self.is_dirty = True
        del self.__db_modules[module.db_id]
    def db_get_module(self, key):
        if self.__db_modules.has_key(key):
            return self.__db_modules[key]
        return None
    
    def __get_db_connections(self):
        return self.__db_connections
    def __set_db_connections(self, connections):
        self.__db_connections = connections
        self.is_dirty = True
    db_connections = property(__get_db_connections, __set_db_connections)
    def db_get_connections(self):
        return self.__db_connections.values()
    def db_add_connection(self, connection):
        self.is_dirty = True
        self.__db_connections[connection.db_id] = connection
    def db_change_connection(self, connection):
        self.is_dirty = True
        self.__db_connections[connection.db_id] = connection
    def db_delete_connection(self, connection):
        self.is_dirty = True
        del self.__db_connections[connection.db_id]
    def db_get_connection(self, key):
        if self.__db_connections.has_key(key):
            return self.__db_connections[key]
        return None
    
    def __get_db_annotations(self):
        return self.__db_annotations
    def __set_db_annotations(self, annotations):
        self.__db_annotations = annotations
        self.is_dirty = True
    db_annotations = property(__get_db_annotations, __set_db_annotations)
    def db_get_annotations(self):
        return self.__db_annotations
    def db_add_annotation(self, annotation):
        self.is_dirty = True
        self.__db_annotations.append(annotation)
    def db_change_annotation(self, annotation):
        self.is_dirty = True
        found = False
        for i in xrange(len(self.__db_annotations)):
            if self.__db_annotations[i].db_id == annotation.db_id:
                self.__db_annotations[i] = annotation
                found = True
                break
        if not found:
            self.__db_annotations.append(annotation)
    def db_delete_annotation(self, annotation):
        self.is_dirty = True
        for i in xrange(len(self.__db_annotations)):
            if self.__db_annotations[i].db_id == annotation.db_id:
                del self.__db_annotations[i]
                break
    def db_get_annotation(self, key):
        for i in xrange(len(self.__db_annotations)):
            if self.__db_annotations[i].db_id == key:
                return self.__db_annotations[i]
        return None
    
    def __get_db_others(self):
        return self.__db_others
    def __set_db_others(self, others):
        self.__db_others = others
        self.is_dirty = True
    db_others = property(__get_db_others, __set_db_others)
    def db_get_others(self):
        return self.__db_others
    def db_add_other(self, other):
        self.is_dirty = True
        self.__db_others.append(other)
    def db_change_other(self, other):
        self.is_dirty = True
        found = False
        for i in xrange(len(self.__db_others)):
            if self.__db_others[i].db_id == other.db_id:
                self.__db_others[i] = other
                found = True
                break
        if not found:
            self.__db_others.append(other)
    def db_delete_other(self, other):
        self.is_dirty = True
        for i in xrange(len(self.__db_others)):
            if self.__db_others[i].db_id == other.db_id:
                del self.__db_others[i]
                break
    def db_get_other(self, key):
        for i in xrange(len(self.__db_others)):
            if self.__db_others[i].db_id == key:
                return self.__db_others[i]
        return None
    
    def getPrimaryKey(self):
        return self.__db_id

class DBAction(object):

    vtType = 'action'

    def __init__(self, id=None, prevId=None, date=None, user=None, operations=None):
        self.__db_id = id
        self.__db_prevId = prevId
        self.__db_date = date
        self.__db_user = user
        if operations is None:
            self.__db_operations = []
        else:
            self.__db_operations = operations
        self.is_dirty = True
    
    def __copy__(self):
        cp = DBAction()
        cp.db_id = self.db_id
        cp.is_dirty = self.is_dirty
        cp.db_prevId = self.db_prevId
        cp.is_dirty = self.is_dirty
        cp.db_date = self.db_date
        cp.is_dirty = self.is_dirty
        cp.db_user = self.db_user
        cp.is_dirty = self.is_dirty
        if self.db_operations is None:
            cp.db_operations = None
        else:
            cp.db_operations = [copy.copy(v) for v in self.db_operations]
        cp.is_dirty = self.is_dirty
        return cp

    def db_children(self, parent=(None,None), orphan=False):
        children = []
        children.append((self, parent[0], parent[1]))
        return children
    def __get_db_id(self):
        return self.__db_id
    def __set_db_id(self, id):
        self.__db_id = id
        self.is_dirty = True
    db_id = property(__get_db_id, __set_db_id)
    def db_add_id(self, id):
        self.__db_id = id
    def db_change_id(self, id):
        self.__db_id = id
    def db_delete_id(self, id):
        self.__db_id = None
    
    def __get_db_prevId(self):
        return self.__db_prevId
    def __set_db_prevId(self, prevId):
        self.__db_prevId = prevId
        self.is_dirty = True
    db_prevId = property(__get_db_prevId, __set_db_prevId)
    def db_add_prevId(self, prevId):
        self.__db_prevId = prevId
    def db_change_prevId(self, prevId):
        self.__db_prevId = prevId
    def db_delete_prevId(self, prevId):
        self.__db_prevId = None
    
    def __get_db_date(self):
        return self.__db_date
    def __set_db_date(self, date):
        self.__db_date = date
        self.is_dirty = True
    db_date = property(__get_db_date, __set_db_date)
    def db_add_date(self, date):
        self.__db_date = date
    def db_change_date(self, date):
        self.__db_date = date
    def db_delete_date(self, date):
        self.__db_date = None
    
    def __get_db_user(self):
        return self.__db_user
    def __set_db_user(self, user):
        self.__db_user = user
        self.is_dirty = True
    db_user = property(__get_db_user, __set_db_user)
    def db_add_user(self, user):
        self.__db_user = user
    def db_change_user(self, user):
        self.__db_user = user
    def db_delete_user(self, user):
        self.__db_user = None
    
    def __get_db_operations(self):
        return self.__db_operations
    def __set_db_operations(self, operations):
        self.__db_operations = operations
        self.is_dirty = True
    db_operations = property(__get_db_operations, __set_db_operations)
    def db_get_operations(self):
        return self.__db_operations
    def db_add_operation(self, operation):
        self.is_dirty = True
        self.__db_operations.append(operation)
    def db_change_operation(self, operation):
        self.is_dirty = True
        found = False
        for i in xrange(len(self.__db_operations)):
            if self.__db_operations[i].db_id == operation.db_id:
                self.__db_operations[i] = operation
                found = True
                break
        if not found:
            self.__db_operations.append(operation)
    def db_delete_operation(self, operation):
        self.is_dirty = True
        for i in xrange(len(self.__db_operations)):
            if self.__db_operations[i].db_id == operation.db_id:
                del self.__db_operations[i]
                break
    def db_get_operation(self, key):
        for i in xrange(len(self.__db_operations)):
            if self.__db_operations[i].db_id == key:
                return self.__db_operations[i]
        return None
    
    def getPrimaryKey(self):
        return self.__db_id

class DBAnnotation(object):

    vtType = 'annotation'

    def __init__(self, id=None, key=None, value=None):
        self.__db_id = id
        self.__db_key = key
        self.__db_value = value
        self.is_dirty = True
    
    def __copy__(self):
        cp = DBAnnotation()
        cp.db_id = self.db_id
        cp.is_dirty = self.is_dirty
        cp.db_key = self.db_key
        cp.is_dirty = self.is_dirty
        cp.db_value = self.db_value
        cp.is_dirty = self.is_dirty
        return cp

    def db_children(self, parent=(None,None), orphan=False):
        children = []
        children.append((self, parent[0], parent[1]))
        return children
    def __get_db_id(self):
        return self.__db_id
    def __set_db_id(self, id):
        self.__db_id = id
        self.is_dirty = True
    db_id = property(__get_db_id, __set_db_id)
    def db_add_id(self, id):
        self.__db_id = id
    def db_change_id(self, id):
        self.__db_id = id
    def db_delete_id(self, id):
        self.__db_id = None
    
    def __get_db_key(self):
        return self.__db_key
    def __set_db_key(self, key):
        self.__db_key = key
        self.is_dirty = True
    db_key = property(__get_db_key, __set_db_key)
    def db_add_key(self, key):
        self.__db_key = key
    def db_change_key(self, key):
        self.__db_key = key
    def db_delete_key(self, key):
        self.__db_key = None
    
    def __get_db_value(self):
        return self.__db_value
    def __set_db_value(self, value):
        self.__db_value = value
        self.is_dirty = True
    db_value = property(__get_db_value, __set_db_value)
    def db_add_value(self, value):
        self.__db_value = value
    def db_change_value(self, value):
        self.__db_value = value
    def db_delete_value(self, value):
        self.__db_value = None
    
    def getPrimaryKey(self):
        return self.__db_id

class DBChange(object):

    vtType = 'change'

    def __init__(self, id=None, what=None, oldObjId=None, newObjId=None, parentObjId=None, parentObjType=None, data=None):
        self.__db_id = id
        self.__db_what = what
        self.__db_oldObjId = oldObjId
        self.__db_newObjId = newObjId
        self.__db_parentObjId = parentObjId
        self.__db_parentObjType = parentObjType
        self.__db_data = data
        self.is_dirty = True
    
    def __copy__(self):
        cp = DBChange()
        cp.db_id = self.db_id
        cp.is_dirty = self.is_dirty
        cp.db_what = self.db_what
        cp.is_dirty = self.is_dirty
        cp.db_oldObjId = self.db_oldObjId
        cp.is_dirty = self.is_dirty
        cp.db_newObjId = self.db_newObjId
        cp.is_dirty = self.is_dirty
        cp.db_parentObjId = self.db_parentObjId
        cp.is_dirty = self.is_dirty
        cp.db_parentObjType = self.db_parentObjType
        cp.is_dirty = self.is_dirty
        cp.db_data = self.db_data
        cp.is_dirty = self.is_dirty
        return cp

    def db_children(self, parent=(None,None), orphan=False):
        children = []
        children.extend(self.db_action.db_children((self.vtType, self.db_id), orphan))
        if orphan:
            self.db_action = None
        children.append((self, parent[0], parent[1]))
        return children
    def __get_db_id(self):
        return self.__db_id
    def __set_db_id(self, id):
        self.__db_id = id
        self.is_dirty = True
    db_id = property(__get_db_id, __set_db_id)
    def db_add_id(self, id):
        self.__db_id = id
    def db_change_id(self, id):
        self.__db_id = id
    def db_delete_id(self, id):
        self.__db_id = None
    
    def __get_db_what(self):
        return self.__db_what
    def __set_db_what(self, what):
        self.__db_what = what
        self.is_dirty = True
    db_what = property(__get_db_what, __set_db_what)
    def db_add_what(self, what):
        self.__db_what = what
    def db_change_what(self, what):
        self.__db_what = what
    def db_delete_what(self, what):
        self.__db_what = None
    
    def __get_db_oldObjId(self):
        return self.__db_oldObjId
    def __set_db_oldObjId(self, oldObjId):
        self.__db_oldObjId = oldObjId
        self.is_dirty = True
    db_oldObjId = property(__get_db_oldObjId, __set_db_oldObjId)
    def db_add_oldObjId(self, oldObjId):
        self.__db_oldObjId = oldObjId
    def db_change_oldObjId(self, oldObjId):
        self.__db_oldObjId = oldObjId
    def db_delete_oldObjId(self, oldObjId):
        self.__db_oldObjId = None
    
    def __get_db_newObjId(self):
        return self.__db_newObjId
    def __set_db_newObjId(self, newObjId):
        self.__db_newObjId = newObjId
        self.is_dirty = True
    db_newObjId = property(__get_db_newObjId, __set_db_newObjId)
    def db_add_newObjId(self, newObjId):
        self.__db_newObjId = newObjId
    def db_change_newObjId(self, newObjId):
        self.__db_newObjId = newObjId
    def db_delete_newObjId(self, newObjId):
        self.__db_newObjId = None
    
    def __get_db_parentObjId(self):
        return self.__db_parentObjId
    def __set_db_parentObjId(self, parentObjId):
        self.__db_parentObjId = parentObjId
        self.is_dirty = True
    db_parentObjId = property(__get_db_parentObjId, __set_db_parentObjId)
    def db_add_parentObjId(self, parentObjId):
        self.__db_parentObjId = parentObjId
    def db_change_parentObjId(self, parentObjId):
        self.__db_parentObjId = parentObjId
    def db_delete_parentObjId(self, parentObjId):
        self.__db_parentObjId = None
    
    def __get_db_parentObjType(self):
        return self.__db_parentObjType
    def __set_db_parentObjType(self, parentObjType):
        self.__db_parentObjType = parentObjType
        self.is_dirty = True
    db_parentObjType = property(__get_db_parentObjType, __set_db_parentObjType)
    def db_add_parentObjType(self, parentObjType):
        self.__db_parentObjType = parentObjType
    def db_change_parentObjType(self, parentObjType):
        self.__db_parentObjType = parentObjType
    def db_delete_parentObjType(self, parentObjType):
        self.__db_parentObjType = None
    
    def __get_db_data(self):
        return self.__db_data
    def __set_db_data(self, data):
        self.__db_data = data
        self.is_dirty = True
    db_data = property(__get_db_data, __set_db_data)
    def db_add_data(self, data):
        self.__db_data = data
    def db_change_data(self, data):
        self.__db_data = data
    def db_delete_data(self, data):
        self.__db_data = None
    
    def getPrimaryKey(self):
        return self.__db_id

class DBMacro(object):

    vtType = 'macro'

    def __init__(self, id=None, name=None, descrptn=None, actions=None):
        self.__db_id = id
        self.__db_name = name
        self.__db_descrptn = descrptn
        if actions is None:
            self.__db_actions = {}
        else:
            self.__db_actions = actions
        self.is_dirty = True
    
    def __copy__(self):
        cp = DBMacro()
        cp.db_id = self.db_id
        cp.is_dirty = self.is_dirty
        cp.db_name = self.db_name
        cp.is_dirty = self.is_dirty
        cp.db_descrptn = self.db_descrptn
        cp.is_dirty = self.is_dirty
        if self.db_actions is None:
            cp.db_actions = None
        else:
            cp.db_actions = dict([(k,copy.copy(v)) for (k,v) in self.db_actions.iteritems()])
        cp.is_dirty = self.is_dirty
        return cp

    def db_children(self, parent=(None,None), orphan=False):
        children = []
        for child in self.db_actions.itervalues():
            children.extend(child.db_children((self.vtType, self.db_id), orphan))
        if orphan:
            self.db_actions = {}
        children.extend(self.db_vistrail.db_children((self.vtType, self.db_id), orphan))
        if orphan:
            self.db_vistrail = None
        children.append((self, parent[0], parent[1]))
        return children
    def __get_db_id(self):
        return self.__db_id
    def __set_db_id(self, id):
        self.__db_id = id
        self.is_dirty = True
    db_id = property(__get_db_id, __set_db_id)
    def db_add_id(self, id):
        self.__db_id = id
    def db_change_id(self, id):
        self.__db_id = id
    def db_delete_id(self, id):
        self.__db_id = None
    
    def __get_db_name(self):
        return self.__db_name
    def __set_db_name(self, name):
        self.__db_name = name
        self.is_dirty = True
    db_name = property(__get_db_name, __set_db_name)
    def db_add_name(self, name):
        self.__db_name = name
    def db_change_name(self, name):
        self.__db_name = name
    def db_delete_name(self, name):
        self.__db_name = None
    
    def __get_db_descrptn(self):
        return self.__db_descrptn
    def __set_db_descrptn(self, descrptn):
        self.__db_descrptn = descrptn
        self.is_dirty = True
    db_descrptn = property(__get_db_descrptn, __set_db_descrptn)
    def db_add_descrptn(self, descrptn):
        self.__db_descrptn = descrptn
    def db_change_descrptn(self, descrptn):
        self.__db_descrptn = descrptn
    def db_delete_descrptn(self, descrptn):
        self.__db_descrptn = None
    
    def __get_db_actions(self):
        return self.__db_actions
    def __set_db_actions(self, actions):
        self.__db_actions = actions
        self.is_dirty = True
    db_actions = property(__get_db_actions, __set_db_actions)
    def db_get_actions(self):
        return self.__db_actions.values()
    def db_add_action(self, action):
        self.is_dirty = True
        self.__db_actions[action.db_id] = action
    def db_change_action(self, action):
        self.is_dirty = True
        self.__db_actions[action.db_id] = action
    def db_delete_action(self, action):
        self.is_dirty = True
        del self.__db_actions[action.db_id]
    def db_get_action(self, key):
        if self.__db_actions.has_key(key):
            return self.__db_actions[key]
        return None
    
    def getPrimaryKey(self):
        return self.__db_id

class DBConnection(object):

    vtType = 'connection'

    def __init__(self, id=None, ports=None):
        self.__db_id = id
        self.db_type_index = {}
        if ports is None:
            self.__db_ports = []
        else:
            self.__db_ports = ports
            for v in self.__db_ports:
                self.db_type_index[v.db_type] = v
        self.is_dirty = True
    
    def __copy__(self):
        cp = DBConnection()
        cp.db_id = self.db_id
        cp.is_dirty = self.is_dirty
        if self.db_ports is None:
            cp.db_ports = None
        else:
            cp.db_ports = [copy.copy(v) for v in self.db_ports]
            for v in cp.__db_ports:
                cp.db_type_index[v.db_type] = v
        cp.is_dirty = self.is_dirty
        return cp

    def db_children(self, parent=(None,None), orphan=False):
        children = []
        for child in self.db_ports:
            children.extend(child.db_children((self.vtType, self.db_id), orphan))
        if orphan:
            self.db_ports = []
        children.append((self, parent[0], parent[1]))
        return children
    def __get_db_id(self):
        return self.__db_id
    def __set_db_id(self, id):
        self.__db_id = id
        self.is_dirty = True
    db_id = property(__get_db_id, __set_db_id)
    def db_add_id(self, id):
        self.__db_id = id
    def db_change_id(self, id):
        self.__db_id = id
    def db_delete_id(self, id):
        self.__db_id = None
    
    def __get_db_ports(self):
        return self.__db_ports
    def __set_db_ports(self, ports):
        self.__db_ports = ports
        self.is_dirty = True
    db_ports = property(__get_db_ports, __set_db_ports)
    def db_get_ports(self):
        return self.__db_ports
    def db_add_port(self, port):
        self.is_dirty = True
        self.__db_ports.append(port)
        self.db_type_index[port.db_type] = port
    def db_change_port(self, port):
        self.is_dirty = True
        found = False
        for i in xrange(len(self.__db_ports)):
            if self.__db_ports[i].db_id == port.db_id:
                self.__db_ports[i] = port
                found = True
                break
        if not found:
            self.__db_ports.append(port)
        self.db_type_index[port.db_type] = port
    def db_delete_port(self, port):
        self.is_dirty = True
        for i in xrange(len(self.__db_ports)):
            if self.__db_ports[i].db_id == port.db_id:
                del self.__db_ports[i]
                break
        del self.db_type_index[port.db_type]
    def db_get_port(self, key):
        for i in xrange(len(self.__db_ports)):
            if self.__db_ports[i].db_id == key:
                return self.__db_ports[i]
        return None
    def db_get_port_by_type(self, key):
        return self.db_type_index[key]
    
    def getPrimaryKey(self):
        return self.__db_id

class DBTag(object):

    vtType = 'tag'

    def __init__(self, name=None, time=None):
        self.__db_name = name
        self.__db_time = time
        self.is_dirty = True
    
    def __copy__(self):
        cp = DBTag()
        cp.db_name = self.db_name
        cp.is_dirty = self.is_dirty
        cp.db_time = self.db_time
        cp.is_dirty = self.is_dirty
        return cp

    def db_children(self, parent=(None,None), orphan=False):
        children = []
        children.extend(self.db_vistrail.db_children((self.vtType, self.db_id), orphan))
        if orphan:
            self.db_vistrail = None
        children.append((self, parent[0], parent[1]))
        return children
    def __get_db_name(self):
        return self.__db_name
    def __set_db_name(self, name):
        self.__db_name = name
        self.is_dirty = True
    db_name = property(__get_db_name, __set_db_name)
    def db_add_name(self, name):
        self.__db_name = name
    def db_change_name(self, name):
        self.__db_name = name
    def db_delete_name(self, name):
        self.__db_name = None
    
    def __get_db_time(self):
        return self.__db_time
    def __set_db_time(self, time):
        self.__db_time = time
        self.is_dirty = True
    db_time = property(__get_db_time, __set_db_time)
    def db_add_time(self, time):
        self.__db_time = time
    def db_change_time(self, time):
        self.__db_time = time
    def db_delete_time(self, time):
        self.__db_time = None
    
    def getPrimaryKey(self):
        return self.__db_name

class DBExecRec(object):

    vtType = 'execRec'

    def __init__(self, id=None, tsStart=None, tsEnd=None, moduleId=None, moduleName=None, annotations=None):
        self.__db_id = id
        self.__db_tsStart = tsStart
        self.__db_tsEnd = tsEnd
        self.__db_moduleId = moduleId
        self.__db_moduleName = moduleName
        if annotations is None:
            self.__db_annotations = []
        else:
            self.__db_annotations = annotations
        self.is_dirty = True
    
    def __copy__(self):
        cp = DBExecRec()
        cp.db_id = self.db_id
        cp.is_dirty = self.is_dirty
        cp.db_tsStart = self.db_tsStart
        cp.is_dirty = self.is_dirty
        cp.db_tsEnd = self.db_tsEnd
        cp.is_dirty = self.is_dirty
        cp.db_moduleId = self.db_moduleId
        cp.is_dirty = self.is_dirty
        cp.db_moduleName = self.db_moduleName
        cp.is_dirty = self.is_dirty
        if self.db_annotations is None:
            cp.db_annotations = None
        else:
            cp.db_annotations = [copy.copy(v) for v in self.db_annotations]
        cp.is_dirty = self.is_dirty
        return cp

    def db_children(self, parent=(None,None), orphan=False):
        children = []
        for child in self.db_annotations:
            children.extend(child.db_children((self.vtType, self.db_id), orphan))
        if orphan:
            self.db_annotations = []
        children.extend(self.db_wfExec.db_children((self.vtType, self.db_id), orphan))
        if orphan:
            self.db_wfExec = None
        children.append((self, parent[0], parent[1]))
        return children
    def __get_db_id(self):
        return self.__db_id
    def __set_db_id(self, id):
        self.__db_id = id
        self.is_dirty = True
    db_id = property(__get_db_id, __set_db_id)
    def db_add_id(self, id):
        self.__db_id = id
    def db_change_id(self, id):
        self.__db_id = id
    def db_delete_id(self, id):
        self.__db_id = None
    
    def __get_db_tsStart(self):
        return self.__db_tsStart
    def __set_db_tsStart(self, tsStart):
        self.__db_tsStart = tsStart
        self.is_dirty = True
    db_tsStart = property(__get_db_tsStart, __set_db_tsStart)
    def db_add_tsStart(self, tsStart):
        self.__db_tsStart = tsStart
    def db_change_tsStart(self, tsStart):
        self.__db_tsStart = tsStart
    def db_delete_tsStart(self, tsStart):
        self.__db_tsStart = None
    
    def __get_db_tsEnd(self):
        return self.__db_tsEnd
    def __set_db_tsEnd(self, tsEnd):
        self.__db_tsEnd = tsEnd
        self.is_dirty = True
    db_tsEnd = property(__get_db_tsEnd, __set_db_tsEnd)
    def db_add_tsEnd(self, tsEnd):
        self.__db_tsEnd = tsEnd
    def db_change_tsEnd(self, tsEnd):
        self.__db_tsEnd = tsEnd
    def db_delete_tsEnd(self, tsEnd):
        self.__db_tsEnd = None
    
    def __get_db_moduleId(self):
        return self.__db_moduleId
    def __set_db_moduleId(self, moduleId):
        self.__db_moduleId = moduleId
        self.is_dirty = True
    db_moduleId = property(__get_db_moduleId, __set_db_moduleId)
    def db_add_moduleId(self, moduleId):
        self.__db_moduleId = moduleId
    def db_change_moduleId(self, moduleId):
        self.__db_moduleId = moduleId
    def db_delete_moduleId(self, moduleId):
        self.__db_moduleId = None
    
    def __get_db_moduleName(self):
        return self.__db_moduleName
    def __set_db_moduleName(self, moduleName):
        self.__db_moduleName = moduleName
        self.is_dirty = True
    db_moduleName = property(__get_db_moduleName, __set_db_moduleName)
    def db_add_moduleName(self, moduleName):
        self.__db_moduleName = moduleName
    def db_change_moduleName(self, moduleName):
        self.__db_moduleName = moduleName
    def db_delete_moduleName(self, moduleName):
        self.__db_moduleName = None
    
    def __get_db_annotations(self):
        return self.__db_annotations
    def __set_db_annotations(self, annotations):
        self.__db_annotations = annotations
        self.is_dirty = True
    db_annotations = property(__get_db_annotations, __set_db_annotations)
    def db_get_annotations(self):
        return self.__db_annotations
    def db_add_annotation(self, annotation):
        self.is_dirty = True
        self.__db_annotations.append(annotation)
    def db_change_annotation(self, annotation):
        self.is_dirty = True
        found = False
        for i in xrange(len(self.__db_annotations)):
            if self.__db_annotations[i].db_id == annotation.db_id:
                self.__db_annotations[i] = annotation
                found = True
                break
        if not found:
            self.__db_annotations.append(annotation)
    def db_delete_annotation(self, annotation):
        self.is_dirty = True
        for i in xrange(len(self.__db_annotations)):
            if self.__db_annotations[i].db_id == annotation.db_id:
                del self.__db_annotations[i]
                break
    def db_get_annotation(self, key):
        for i in xrange(len(self.__db_annotations)):
            if self.__db_annotations[i].db_id == key:
                return self.__db_annotations[i]
        return None
    
    def getPrimaryKey(self):
        return self.__db_id

class DBVistrail(object):

    vtType = 'vistrail'

    def __init__(self, id=None, version=None, name=None, dbHost=None, dbPort=None, dbName=None, actions=None, tags=None, macros=None):
        self.__db_id = id
        self.__db_version = version
        self.__db_name = name
        self.__db_dbHost = dbHost
        self.__db_dbPort = dbPort
        self.__db_dbName = dbName
        if actions is None:
            self.__db_actions = {}
        else:
            self.__db_actions = actions
        if tags is None:
            self.__db_tags = {}
        else:
            self.__db_tags = tags
        if macros is None:
            self.__db_macros = {}
        else:
            self.__db_macros = macros
        self.is_dirty = True
    
    def __copy__(self):
        cp = DBVistrail()
        cp.db_id = self.db_id
        cp.is_dirty = self.is_dirty
        cp.db_version = self.db_version
        cp.is_dirty = self.is_dirty
        cp.db_name = self.db_name
        cp.is_dirty = self.is_dirty
        cp.db_dbHost = self.db_dbHost
        cp.is_dirty = self.is_dirty
        cp.db_dbPort = self.db_dbPort
        cp.is_dirty = self.is_dirty
        cp.db_dbName = self.db_dbName
        cp.is_dirty = self.is_dirty
        if self.db_actions is None:
            cp.db_actions = None
        else:
            cp.db_actions = dict([(k,copy.copy(v)) for (k,v) in self.db_actions.iteritems()])
        cp.is_dirty = self.is_dirty
        if self.db_tags is None:
            cp.db_tags = None
        else:
            cp.db_tags = dict([(k,copy.copy(v)) for (k,v) in self.db_tags.iteritems()])
        cp.is_dirty = self.is_dirty
        if self.db_macros is None:
            cp.db_macros = None
        else:
            cp.db_macros = dict([(k,copy.copy(v)) for (k,v) in self.db_macros.iteritems()])
        cp.is_dirty = self.is_dirty
        return cp

    def db_children(self, parent=(None,None), orphan=False):
        children = []
        for child in self.db_actions.itervalues():
            children.extend(child.db_children((self.vtType, self.db_id), orphan))
        if orphan:
            self.db_actions = {}
        for child in self.db_tags.itervalues():
            children.extend(child.db_children((self.vtType, self.db_id), orphan))
        if orphan:
            self.db_tags = {}
        for child in self.db_macros.itervalues():
            children.extend(child.db_children((self.vtType, self.db_id), orphan))
        if orphan:
            self.db_macros = {}
        children.append((self, parent[0], parent[1]))
        return children
    def __get_db_id(self):
        return self.__db_id
    def __set_db_id(self, id):
        self.__db_id = id
        self.is_dirty = True
    db_id = property(__get_db_id, __set_db_id)
    def db_add_id(self, id):
        self.__db_id = id
    def db_change_id(self, id):
        self.__db_id = id
    def db_delete_id(self, id):
        self.__db_id = None
    
    def __get_db_version(self):
        return self.__db_version
    def __set_db_version(self, version):
        self.__db_version = version
        self.is_dirty = True
    db_version = property(__get_db_version, __set_db_version)
    def db_add_version(self, version):
        self.__db_version = version
    def db_change_version(self, version):
        self.__db_version = version
    def db_delete_version(self, version):
        self.__db_version = None
    
    def __get_db_name(self):
        return self.__db_name
    def __set_db_name(self, name):
        self.__db_name = name
        self.is_dirty = True
    db_name = property(__get_db_name, __set_db_name)
    def db_add_name(self, name):
        self.__db_name = name
    def db_change_name(self, name):
        self.__db_name = name
    def db_delete_name(self, name):
        self.__db_name = None
    
    def __get_db_dbHost(self):
        return self.__db_dbHost
    def __set_db_dbHost(self, dbHost):
        self.__db_dbHost = dbHost
        self.is_dirty = True
    db_dbHost = property(__get_db_dbHost, __set_db_dbHost)
    def db_add_dbHost(self, dbHost):
        self.__db_dbHost = dbHost
    def db_change_dbHost(self, dbHost):
        self.__db_dbHost = dbHost
    def db_delete_dbHost(self, dbHost):
        self.__db_dbHost = None
    
    def __get_db_dbPort(self):
        return self.__db_dbPort
    def __set_db_dbPort(self, dbPort):
        self.__db_dbPort = dbPort
        self.is_dirty = True
    db_dbPort = property(__get_db_dbPort, __set_db_dbPort)
    def db_add_dbPort(self, dbPort):
        self.__db_dbPort = dbPort
    def db_change_dbPort(self, dbPort):
        self.__db_dbPort = dbPort
    def db_delete_dbPort(self, dbPort):
        self.__db_dbPort = None
    
    def __get_db_dbName(self):
        return self.__db_dbName
    def __set_db_dbName(self, dbName):
        self.__db_dbName = dbName
        self.is_dirty = True
    db_dbName = property(__get_db_dbName, __set_db_dbName)
    def db_add_dbName(self, dbName):
        self.__db_dbName = dbName
    def db_change_dbName(self, dbName):
        self.__db_dbName = dbName
    def db_delete_dbName(self, dbName):
        self.__db_dbName = None
    
    def __get_db_actions(self):
        return self.__db_actions
    def __set_db_actions(self, actions):
        self.__db_actions = actions
        self.is_dirty = True
    db_actions = property(__get_db_actions, __set_db_actions)
    def db_get_actions(self):
        return self.__db_actions.values()
    def db_add_action(self, action):
        self.is_dirty = True
        self.__db_actions[action.db_id] = action
    def db_change_action(self, action):
        self.is_dirty = True
        self.__db_actions[action.db_id] = action
    def db_delete_action(self, action):
        self.is_dirty = True
        del self.__db_actions[action.db_id]
    def db_get_action(self, key):
        if self.__db_actions.has_key(key):
            return self.__db_actions[key]
        return None
    
    def __get_db_tags(self):
        return self.__db_tags
    def __set_db_tags(self, tags):
        self.__db_tags = tags
        self.is_dirty = True
    db_tags = property(__get_db_tags, __set_db_tags)
    def db_get_tags(self):
        return self.__db_tags.values()
    def db_add_tag(self, tag):
        self.is_dirty = True
        self.__db_tags[tag.db_name] = tag
    def db_change_tag(self, tag):
        self.is_dirty = True
        self.__db_tags[tag.db_name] = tag
    def db_delete_tag(self, tag):
        self.is_dirty = True
        del self.__db_tags[tag.db_name]
    def db_get_tag(self, key):
        if self.__db_tags.has_key(key):
            return self.__db_tags[key]
        return None
    
    def __get_db_macros(self):
        return self.__db_macros
    def __set_db_macros(self, macros):
        self.__db_macros = macros
        self.is_dirty = True
    db_macros = property(__get_db_macros, __set_db_macros)
    def db_get_macros(self):
        return self.__db_macros.values()
    def db_add_macro(self, macro):
        self.is_dirty = True
        self.__db_macros[macro.db_id] = macro
    def db_change_macro(self, macro):
        self.is_dirty = True
        self.__db_macros[macro.db_id] = macro
    def db_delete_macro(self, macro):
        self.is_dirty = True
        del self.__db_macros[macro.db_id]
    def db_get_macro(self, key):
        if self.__db_macros.has_key(key):
            return self.__db_macros[key]
        return None
    
    def getPrimaryKey(self):
        return self.__db_id

class DBDelete(object):

    vtType = 'delete'

    def __init__(self, id=None, what=None, objectId=None, parentObjId=None, parentObjType=None):
        self.__db_id = id
        self.__db_what = what
        self.__db_objectId = objectId
        self.__db_parentObjId = parentObjId
        self.__db_parentObjType = parentObjType
        self.is_dirty = True
    
    def __copy__(self):
        cp = DBDelete()
        cp.db_id = self.db_id
        cp.is_dirty = self.is_dirty
        cp.db_what = self.db_what
        cp.is_dirty = self.is_dirty
        cp.db_objectId = self.db_objectId
        cp.is_dirty = self.is_dirty
        cp.db_parentObjId = self.db_parentObjId
        cp.is_dirty = self.is_dirty
        cp.db_parentObjType = self.db_parentObjType
        cp.is_dirty = self.is_dirty
        return cp

    def db_children(self, parent=(None,None), orphan=False):
        children = []
        children.extend(self.db_action.db_children((self.vtType, self.db_id), orphan))
        if orphan:
            self.db_action = None
        children.append((self, parent[0], parent[1]))
        return children
    def __get_db_id(self):
        return self.__db_id
    def __set_db_id(self, id):
        self.__db_id = id
        self.is_dirty = True
    db_id = property(__get_db_id, __set_db_id)
    def db_add_id(self, id):
        self.__db_id = id
    def db_change_id(self, id):
        self.__db_id = id
    def db_delete_id(self, id):
        self.__db_id = None
    
    def __get_db_what(self):
        return self.__db_what
    def __set_db_what(self, what):
        self.__db_what = what
        self.is_dirty = True
    db_what = property(__get_db_what, __set_db_what)
    def db_add_what(self, what):
        self.__db_what = what
    def db_change_what(self, what):
        self.__db_what = what
    def db_delete_what(self, what):
        self.__db_what = None
    
    def __get_db_objectId(self):
        return self.__db_objectId
    def __set_db_objectId(self, objectId):
        self.__db_objectId = objectId
        self.is_dirty = True
    db_objectId = property(__get_db_objectId, __set_db_objectId)
    def db_add_objectId(self, objectId):
        self.__db_objectId = objectId
    def db_change_objectId(self, objectId):
        self.__db_objectId = objectId
    def db_delete_objectId(self, objectId):
        self.__db_objectId = None
    
    def __get_db_parentObjId(self):
        return self.__db_parentObjId
    def __set_db_parentObjId(self, parentObjId):
        self.__db_parentObjId = parentObjId
        self.is_dirty = True
    db_parentObjId = property(__get_db_parentObjId, __set_db_parentObjId)
    def db_add_parentObjId(self, parentObjId):
        self.__db_parentObjId = parentObjId
    def db_change_parentObjId(self, parentObjId):
        self.__db_parentObjId = parentObjId
    def db_delete_parentObjId(self, parentObjId):
        self.__db_parentObjId = None
    
    def __get_db_parentObjType(self):
        return self.__db_parentObjType
    def __set_db_parentObjType(self, parentObjType):
        self.__db_parentObjType = parentObjType
        self.is_dirty = True
    db_parentObjType = property(__get_db_parentObjType, __set_db_parentObjType)
    def db_add_parentObjType(self, parentObjType):
        self.__db_parentObjType = parentObjType
    def db_change_parentObjType(self, parentObjType):
        self.__db_parentObjType = parentObjType
    def db_delete_parentObjType(self, parentObjType):
        self.__db_parentObjType = None
    
    def getPrimaryKey(self):
        return self.__db_id




###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

from __future__ import division

import matplotlib
matplotlib.use('Qt4Agg', warn=False)

import vistrails.core.modules.module_registry
import vistrails.core.db.action
from vistrails.core.vistrail.module import Module
from vistrails.core.vistrail.operation import AddOp

from bases import _modules as _base_modules, MplFigureOutput
from plots import _modules as _plot_modules
from artists import _modules as _artist_modules
from identifiers import identifier

################################################################################

#list of modules to be displaced on matplotlib.new package
_modules = _base_modules + _plot_modules + _artist_modules

def initialize(*args, **kwargs):
    reg = vistrails.core.modules.module_registry.get_module_registry()
    if reg.has_module('org.vistrails.vistrails.spreadsheet',
                      'SpreadsheetCell'):
        from figure_cell import MplFigureCell, MplFigureToSpreadsheet
        _modules.append(MplFigureCell)
        MplFigureOutput.register_output_mode(MplFigureToSpreadsheet)

def handle_module_upgrade_request(controller, module_id, pipeline):
    from vistrails.core.upgradeworkflow import UpgradeWorkflowHandler
    create_new_connection = UpgradeWorkflowHandler.create_new_connection
    reg = vistrails.core.modules.module_registry.get_module_registry()

    def find_module_in_upgrade_action(action):
        for op in action.operations:
            if isinstance(op, AddOp):
                if op.what == Module.vtType:
                    return op.data
        return None

    def find_inputs(m):
        functions = {}
        for f in m.functions:
            if f.name not in functions:
                functions[f.name] = [f]
            else:
                functions[f.name].append(f)
        connections = {}
        for edge in pipeline.graph.iter_edges_to(m.id):
            c = pipeline.connections[edge[2]]
            if c.destination.name not in connections:
                connections[c.destination.name] = [c]
            else:
                connections[c.destination.name].append(c)
        return (functions, connections)

    def find_figure(m):
        for edge in pipeline.graph.iter_edges_from(m.id):
            to_m = pipeline.modules[edge[1]]
            if to_m.name == 'MplFigure':
                # !!! assume only a single down-stream MplFigure !!!
                # may have old or new module...
                if pipeline.connections[edge[2]].destination.name == 'addPlot':
                    return (to_m, None)
                else:
                    return (to_m, edge)
        return (None, None)

    def attach_inputs(new_module, inputs, selected_inputs):
        conns = []
        for port_name in selected_inputs:
            if port_name in inputs[0]:
                for f in inputs[0][port_name]:
                    if len(f.parameters) > 0:
                        new_param_vals, aliases = zip(*[(p.strValue, p.alias) 
                                                        for p in f.parameters])
                    else:
                        new_param_vals = []
                        aliases = []
                    new_f = controller.create_function(new_module, 
                                                       port_name,
                                                       new_param_vals,
                                                       aliases)
                    new_module.add_function(new_f)
            if port_name in inputs[1]:
                for c in inputs[1][port_name]:
                    source_module = pipeline.modules[c.source.id]
                    new_conn = create_new_connection(controller,
                                                     source_module,
                                                     c.source,
                                                     new_module,
                                                     port_name)
                    conns.append(new_conn)
        return conns

    module = pipeline.modules[module_id]
    to_properties = []
    to_axes = []
    old_figure = (None, None)
    props_name = None
    if module.name == 'MplScatterplot':
        props_name = 'MplPathCollectionProperties'
        props_input = 'pathCollectionProperties'
        to_properties = ['facecolor']
        to_axes = ['title', 'xlabel', 'ylabel']
        inputs = find_inputs(module)
        old_loc = module.location
        old_figure = find_figure(module)
    elif module.name == 'MplHistogram':
        props_name = 'MplRectangleProperties'
        props_input = 'rectangleProperties'
        to_properties = ['facecolor']
        to_axes = ['title', 'xlabel', 'ylabel']
        inputs = find_inputs(module)
        old_loc = module.location
        old_figure = find_figure(module)

    module_remap = {'MplPlot':
                    [(None, '1.0.0', 'MplSource',
                      {'dst_port_remap': {'source': 'source',
                                          'Hide Toolbar': None},
                       'src_port_remap': {'source': 'value',
                                          'self': 'value'}})],
                    'MplFigure':
                    [(None, '1.0.0', None,
                      {'dst_port_remap': {'Script': 'addPlot'},
                       'src_port_remap': {'FigureManager': 'self',
                                          'File': 'file'}})],
                    'MplFigureCell':
                    [(None, '1.0.0', None,
                      {'dst_port_remap': {'FigureManager': 'figure'}})],
                    # we will delete parts of this but add back later
                    'MplScatterplot':
                    [(None, '1.0.0', 'MplScatter',
                      {'dst_port_remap': {'xData': 'x',
                                          'yData': 'y',
                                          'facecolor': None,
                                          'title': None,
                                          'xlabel': None,
                                          'ylabel': None,
                                          'self': 'value'},
                       'src_port_remap': {'source': 'value'}})],
                    'MplHistogram':
                    [(None, '1.0.0', 'MplHist',
                      {'dst_port_remap': {'columnData': 'x',
                                          'bins': 'bins',
                                          'facecolor': None,
                                          'title': None,
                                          'xlabel': None,
                                          'ylabel': None,
                                          'self': 'value'},
                       'src_port_remap': {'source': 'value'}})],
                }

    # '1.0.2' -> '1.0.3' changes 'self' output port to 'value'
    module_remap.setdefault('MplSource', []).append(
                (None, '1.0.3', None, {
                 'src_port_remap': {'self': 'value'}}))
    if module.name in (m.__name__ for m in _plot_modules + _artist_modules):
        module_remap.setdefault(module.name, []).append(
                (None, '1.0.3', None, {
                 'src_port_remap': {'self': 'value'}}))

    action_list = []
    if old_figure[1] is not None and \
       any(p in inputs[0] or p in inputs[1] for p in to_axes):
        # need to remove the edge between plot and figure
        pipeline.graph.delete_edge(*old_figure[1])
        conn = pipeline.connections[old_figure[1][2]]
        action = vistrails.core.db.action.create_action([('delete', conn)])
        action_list.append(action)

    try:
        from vistrails.packages.spreadsheet.init import upgrade_cell_to_output
    except ImportError:
        pass
    else:
        module_remap = upgrade_cell_to_output(
                module_remap, module_id, pipeline,
                'MplFigureCell', 'MplFigureOutput',
                '1.0.5', 'figure')

    normal_actions = UpgradeWorkflowHandler.remap_module(
            controller, module_id, pipeline, module_remap)
    action_list.extend(normal_actions)

    more_ops = []
    if any(p in inputs[0] or p in inputs[1] for p in to_properties):
        # create props module
        if props_name is None:
            raise RuntimeError("properties module needed for unknown module "
                               "%s" % module.name)
        desc = reg.get_descriptor_by_name(identifier, props_name)
        props_module = \
            controller.create_module_from_descriptor(desc,
                                                     old_loc.x + 100,
                                                     old_loc.y + 100)
        more_ops.append(('add', props_module))

        # attach functions/connections
        conns = attach_inputs(props_module, inputs, to_properties)
        more_ops.extend([('add', c) for c in conns])
        
        # attach into pipeline
        new_plot_module = find_module_in_upgrade_action(normal_actions[0])
        assert new_plot_module is not None
        new_conn = create_new_connection(controller,
                                         props_module,
                                         'self',
                                         new_plot_module,
                                         props_input)
        more_ops.append(('add', new_conn))

    if any(p in inputs[0] or p in inputs[1] for p in to_axes):
        # create axes module
        desc = reg.get_descriptor_by_name(identifier, "MplAxesProperties")
        if old_figure[0] is not None:
            old_loc = old_figure[0].location
        axes_module = \
            controller.create_module_from_descriptor(desc,
                                                     old_loc.x + 100,
                                                     old_loc.y + 100)
        more_ops.append(('add', axes_module))

        # attach functions/connections
        conns = attach_inputs(axes_module, inputs, to_axes)
        more_ops.extend([('add', c) for c in conns])
        
        # attach axes properties to new figure
        if old_figure[0] is not None and old_figure[1] is not None:
            # remap figure
            fig_action = UpgradeWorkflowHandler.remap_module(controller,
                                                             old_figure[0].id,
                                                             pipeline,
                                                             module_remap)
            fig_module = find_module_in_upgrade_action(fig_action[0])
            assert fig_module is not None
            # add the removed edge back in
            pipeline.graph.add_edge(*old_figure[1])
            action_list.extend(fig_action)

            new_plot_module = find_module_in_upgrade_action(normal_actions[0])
            assert new_plot_module is not None
            conn = create_new_connection(controller,
                                         new_plot_module,
                                         'self',
                                         fig_module,
                                         'addPlot')
            action = vistrails.core.db.action.create_action([('add', conn)])
            action_list.append(action)
        else:
            fig_module = old_figure[0]
        new_conn = create_new_connection(controller,
                                         axes_module,
                                         'self',
                                         fig_module,
                                         'axesProperties')
        more_ops.append(('add', new_conn))

    return action_list
